#!/usr/bin/env python
# Software License Agreement (BSD License)
#
# Copyright (c) 2016, Kei Okada
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Kei Okada nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

"""
This library implement Minimum Jerk trajectory generation, a.k.a Hoff
& Arbib, described in the documents
http://mplab.ucsd.edu/tutorials/minimumJerk.pdf (you can find copy of
this at http://www.shadmehrlab.org/book/minimumjerk.pdf)

Hoff B, Arbib MA (1992) A model of the effects of speed, accuracy, and
perturbation on visually guided reaching. In: Control of arm movement
in space: neurophysiological and computational approaches
(Caminiti R, Johnson PB, Burnod Y, eds), pp 285-306.

~~~~~~~~~~~~~~~~~~~~~~~~ Min Jerk ~~~~~~~~~~~~~~~~~~~~~~~~
A library for computing minimum jerk trajectory for an arbitrary
set of control points in R2, R3, up to RN space.

  x(t) = a0 + a1*t + a2*t^2 + a3*t^3 + a4*t^4 + a5*t^5
 x'(t) = a1 + 2*a2*t + 3*a3*t^2 + 4*a4*t^3 + 5*a5*t^4
x''(t) = 2*a2 + 6*a3*t + 12*a4*t^2 + 20*a5*t^3

Solve this problem from boundary conditions of x(0), x'(0), x''(0) and
x(T), x'(T), x''(T).

ex. usage:

import numpy
import minjerk
points_array = numpy.array([[1, 2, 3], [4, 4, 4],
                            [6, 4, 6], [2, 5, 6],
                            [5, 6, 7]])
m_coeffs = minjerk.minjerk_coefficients(points_array)
m_curve = minjerk.minjerk_trajectory(m_coeffs, 10)
#  plotting example
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
fig = plt.figure()
ax = fig.gca(projection='3d')
#plot bezier curve
ax.plot(m_curve[:,0], m_curve[:,1], m_curve[:,2], 'r')
#plot specified points
ax.plot(points_array[:,0], points_array[:,1], points_array[:,2], 'g*')
ax.set_title("Minimum Jerk Trajectory")
ax.set_xlabel("X")
ax.set_ylabel("Y")
ax.set_zlabel("Z")
ax.legend(["Minimum Jerk", "Control Points"], loc=2)
plt.show()
"""

import numpy as np


def minjerk_coefficients(points_array, duration_array=None):
    """
    Compute the min-jerk coefficients for a given set for user-supplied control pts

    params:
       points_array: array of user-supplied control points
           numpy.array of size N by k
           N is the number of control points
           k is the number of dimensions for each point
       duration_array: array of user-supplied control duration of ech segment
           numpy.array of size N-1
           N is the number of control points

    returns:
      m_coeffs:  k-dimensional array of N-1 x (6 coefficients + 1 duration of each segment)
           numpy.array of size N-1 by (6+1) by k
    """
    (rows, k) = np.shape(points_array)
    N = rows - 1  # N minus 1 because points array includes x_0
    m_coeffs = np.zeros(shape=(k, N, 7))
    x = points_array[0]
    v = np.zeros(k)
    a = np.zeros(k)
    if duration_array is None:
        duration_array = np.array([1.0] * N)
    assert (
        len(duration_array) == N
    ), "Invalid number of intervals chosen (must be equal to N+1={})".format(N)
    for i in range(0, N):
        gx = points_array[i + 1]
        t = duration_array[i]
        if i == N - 1:
            gv = np.zeros(k)
        else:
            t0 = t
            t1 = duration_array[i + 1]
            d0 = points_array[i + 1] - points_array[i]
            d1 = points_array[i + 2] - points_array[i + 1]
            v0 = d0 / t0
            v1 = d1 / t1
            gv = np.where(
                np.multiply(v0, v1) >= 1e-10, 0.5 * (v0 + v1), np.zeros(k)
            )  # 0 + eps
        ga = np.zeros(k)

        A = (gx - (x + v * t + (a / 2.0) * t * t)) / (t * t * t)
        B = (gv - (v + a * t)) / (t * t)
        C = (ga - a) / t

        a0 = x
        a1 = v
        a2 = a / 2.0
        a3 = 10 * A - 4 * B + 0.5 * C
        a4 = (-15 * A + 7 * B - C) / t
        a5 = (6 * A - 3 * B + 0.5 * C) / (t * t)

        x = gx
        v = gv

        m_coeffs[:, i, 0] = a0
        m_coeffs[:, i, 1] = a1
        m_coeffs[:, i, 2] = a2
        m_coeffs[:, i, 3] = a3
        m_coeffs[:, i, 4] = a4
        m_coeffs[:, i, 5] = a5
        m_coeffs[:, i, 6] = t
    return m_coeffs


def determine_dimensions(trajectory_points):
    # Determine dimensions supplied
    position_flag = True
    velocity_flag = (
        len(trajectory_points[0].velocities) != 0
        and len(trajectory_points[-1].velocities) != 0
    )
    acceleration_flag = (
        len(trajectory_points[0].accelerations) != 0
        and len(trajectory_points[-1].accelerations) != 0
    )
    return {
        "positions": position_flag,
        "velocities": velocity_flag,
        "accelerations": acceleration_flag,
    }


def compute_minjerk_coeff(
    num_joints, trajectory_points, point_duration, dimensions_dict
):
    # Compute Full Minimum Jerk Curve
    num_traj_pts = len(trajectory_points)
    num_traj_dim = sum(dimensions_dict.values())
    num_m_values = len(["a0", "a1", "a2", "a3", "a4", "a5", "tm"])
    m_matrix = np.zeros(
        shape=(num_joints, num_traj_dim, num_traj_pts - 1, num_m_values)
    )
    for jnt in range(num_joints):
        traj_array = np.zeros(shape=(len(trajectory_points), num_traj_dim))
        for idx, point in enumerate(trajectory_points):
            current_point = list()
            current_point.append(point.positions[jnt])
            if dimensions_dict["velocities"]:
                current_point.append(point.velocities[jnt])
            if dimensions_dict["accelerations"]:
                current_point.append(point.accelerations[jnt])
            traj_array[idx, :] = current_point
        m_matrix[jnt, :, :, :] = minjerk_coefficients(traj_array, point_duration)
    return m_matrix


def get_minjerk_point(m_matrix, idx, t, cmd_time, dimensions_dict):
    time_from_start = cmd_time
    num_joints = m_matrix.shape[0]
    positions = [0.0] * num_joints
    velocities = [0.0] * num_joints if dimensions_dict["velocities"] else None
    accelerations = [0.0] * num_joints if dimensions_dict["accelerations"] else None

    for jnt in range(num_joints):
        m_point = minjerk_point(m_matrix[jnt, :, :, :], idx, t)
        # Positions at specified time
        positions[jnt] = m_point[0]
        # Velocities at specified time
        if dimensions_dict["velocities"]:
            velocities[jnt] = m_point[1]
        # Accelerations at specified time
        if dimensions_dict["accelerations"]:
            accelerations[jnt] = m_point[-1]
    return {
        "positions": positions,
        "velocities": velocities,
        "accelerations": accelerations,
        "time_from_start": time_from_start,
    }


def minjerk_trajectory(m_coeffs, num_intervals, duration_array=None):
    """
    Iterpolation of the entire minimum jerk trajectory at once,
    using a specified number of intervals between
    control points (encapsulated by m_coeffs).

    params:
        m_coeffs: N-dimensional array of (6+1) x k  coefficients
            for every control point
            numpy.array of size N by (6 + 1) by k
            N is the number of control points
            k is the number of dimensions for each point
        num_intervals: the number of intervals between
            control points
            int > 0
        duration_array: array of user-supplied control duration of segment
            numpy.array of size N-1
            N is the number of control points

    returns:
        m_curve: positions along the minimum trajectory  in k-dimensions
            numpy.array of size N*num_interval+1  by k
            (the +1 is to include the start position on the curve)
    """
    assert (
        num_intervals > 0
    ), "Invalid number of intervals chosen (must be greater than 0)"
    interval = 1.0 / num_intervals
    (num_axes, num_mpts, _) = np.shape(m_coeffs)
    m_curve = np.zeros((num_mpts * num_intervals + 1, num_axes))
    # Copy out initial point
    m_curve[0, :] = m_coeffs[:, 0, 0]
    if duration_array is None:
        duration_array = np.array([1.0] * num_mpts)
    assert (
        len(duration_array) == num_mpts
    ), "Invalid number of intervals chosen (must be equal to N={})".format(num_mpts)
    for current_mpt in range(num_mpts):
        m_coeff_set = m_coeffs[:, current_mpt, range(7)]
        for iteration, t in enumerate(np.linspace(interval, 1, num_intervals)):
            m_curve[(current_mpt * num_intervals + iteration + 1), :] = (
                _minjerk_trajectory_point(m_coeff_set, t * duration_array[current_mpt])
            )
    return m_curve


def _minjerk_trajectory_point(m_coeff, t):
    """
    Internal convenience function for calculating
    a k-dimensional point defined by the supplied
    minimum jerk coefficients. Finds the point that
    describes the current position along the minimum
    trajectory segment for k dimensions.

    params:
        m_coeff => m0...m3: Four k-dimensional minimum jerk
            coefficients each one is a numpy.array
            of size k by 1, so
            m_coeff is a numpy array of size k by (6+1)
            k is the number of dimensions for each
            coefficient
        t: percentage of time elapsed for this segment
            0 <= int <= 1.0

    returns:
        current position in k dimensions
            numpy.array of size 1 by k
    """
    a0 = m_coeff[:, 0]
    a1 = m_coeff[:, 1]
    a2 = m_coeff[:, 2]
    a3 = m_coeff[:, 3]
    a4 = m_coeff[:, 4]
    a5 = m_coeff[:, 5]
    tm = m_coeff[:, 6]

    t = (
        t * tm
    )  # input t is percentage of time elapsed for this segment, tm is the duration of this segment and to calculate x, v, a , t is the time[s] elapsed for this segment

    # calculate x, v, z at the time percentage  t
    # x=a0+a1*t+a2*t*t+a3*t*t*t+a4*t*t*t*t+a5*t*t*t*t*t;
    x = (
        a0
        + a1 * t
        + a2 * np.power(t, 2)
        + a3 * np.power(t, 3)
        + a4 * np.power(t, 4)
        + a5 * np.power(t, 5)
    )
    # v=a1+2*a2*t+3*a3*t*t+4*a4*t*t*t+5*a5*t*t*t*t;
    (
        a1
        + 2 * a2 * t
        + 3 * a3 * np.power(t, 2)
        + 4 * a4 * np.power(t, 3)
        + 5 * a5 * np.power(t, 4)
    )
    # a=2*a2+6*a3*t+12*a4*t*t+20*a5*t*t*t;
    2 * a2 + 6 * a3 * t + 12 * a4 * np.power(t, 2) + 20 * a5 * np.power(t, 3)

    return x


def minjerk_point(m_coeffs, m_index, t):
    """
    Finds the k values that describe the current
    position along the minjerk trajectory for k dimensions.

    params:
        m_coeffs: k-dimensional array
            for every control point with 6 Minimum Jerk coefficients and a segument duration
            numpy.array of size k by N by 7
            N is the number of control points
            k is the number of dimensions for each point
        m_index: index position out between two of
            the N b_coeffs for this point in time
            int
        t: percentage of time that has passed between
            the two control points
            0 <= int <= 1.0

    returns:
        m_point: current position in k dimensions
            numpy.array of size 1 by k
    """
    if m_index <= 0:
        m_point = m_coeffs[:, 0, 0]
    elif m_index > m_coeffs.shape[1]:
        t = 1
        m_coeff_set = m_coeffs[:, m_coeffs.shape[1] - 1, range(7)]
        m_point = _minjerk_trajectory_point(m_coeff_set, t)
    else:
        t = 0.0 if t < 0.0 else t
        t = 1.0 if t > 1.0 else t
        m_coeff_set = m_coeffs[:, m_index - 1, range(7)]
        m_point = _minjerk_trajectory_point(m_coeff_set, t)
    return m_point



from ..io import log
from grip.robot import BulletRobot
from .end_effector_registry import EndEffectorRegistry


class BulletRobotArm(BulletRobot):
    """A class that specialises a BulletRobot representing a robot arm which can optionally have a gripper as an end-effector"""

    def __init__(
        self, gripper_type: str = None, gripper_max_force: float = 40, **kwargs
    ):
        """BulletRobotArm constructor

        Args:
            gripper_type (str, optional): gripper type name from grip.robot.EndEffectorRegistry to be used as the robot end-effector. Defaults to None, meaning no gripper is created.
            gripper_max_force (float, optional): maximum gripper force. Defaults to 40N.
            urdf_file (str): full path to a robot URDF description
            id (int): if the robot entity already exists in the physics engine, then the id of the robot can be passed instead of its urdf_file
            parent (grip.robot.BulletRobot, optional): This robot may be a sub-part of an already existing parent robot (e.g. a robot hand attached to a robot arm)
            phys_id (int): physics engine identifier (the id of a BulletWorld)
            use_fixed_base (bool): Whether or not to make the object fixed in place. Default: True.
            position (numpy.ndarray): shapebase position of the robot
            orientation (numpy.ndarray): base orientation of the robot as en orientation quaterion following JPL format [x, y, z, w]
            ee_index (int, optional): end-effector link id. Defaults to None. Note: either ee_index or tip_link must be set, but not both.
            tip_link (str, optional): end-effector tip link name. Specifies the end-effector tip link, if ee_index is None, then ee_index = self.get_link_id(tip_link). Defaults to None.
            Note: either ee_index or tip_link must be set, but not both.
            ee_mount_link (str, optional): end-effector mount link name, used to specify up at which link the end-effector has been mounted on the robot.
            joint_names (str, optional): joint names that this robot instance should be controlling (can be a subset of the total number of joints of the robot)
                if not passed as a parameter, or None is given, then this BulletRobot will use all joint names as defined in the urdf description
            has_kinematics (bool): enable or disable forward and inverse kinematics with this robot (this should be usually true, but some robots may not need it)
            max_force (float): maximum generalised force
            sliders (bool): whether or not joint sliders should be added to the debug GUI (only used if grip.robot.BulletWorld was created with phys_opt="gui")
            ee_sliders (bool): whether or not cartesian end-effector sliders should be added to the debug GUI (only used if grip.robot.BulletWorld was created with phys_opt="gui")

        """
        super().__init__(**kwargs)

        self.gripper_type = gripper_type

        log.info(f"Gripper type: {self.gripper_type}")
        self.has_gripper = (
            self.gripper_type is not None
            and self.gripper_type != "none"
            and EndEffectorRegistry.is_registered(self.gripper_type)
        )

        log.info(
            f"Gripper exists: {EndEffectorRegistry.is_registered(self.gripper_type)}"
        )
        self.gripper = None

        if self.has_gripper:
            log.info(
                "Gripper class {}".format(EndEffectorRegistry.get(self.gripper_type))
            )

            self.gripper = EndEffectorRegistry.make(
                self.gripper_type,
                parent=self,
                max_force=gripper_max_force,
                tip_link=self.tip_link,
                urdf_file=None,
                sliders=self._sliders,
            )

            self.gripper.control_fingers(mode="open")

    def state(self) -> dict:
        """Returns the current state of this robot, which includes its joint names, positions, velocities, efforts, end-effector pose and a pointer to self."""

        state = BulletRobot.state(self)

        if self.gripper is not None:
            state["gripper_state"] = self.gripper.state()

        return state

    def reset(self) -> None:
        """Resets robot angles to its home position."""

        super().reset()

        self.gripper.reset()



#!/usr/bin/env python3
import os
import numpy as np
import pybullet as p

import grip
from .template_env import TemplateEnvironment
from .. import robot as g
from ..io import get_data_path


import pybullet_data as pd


base_dir = get_data_path()

TRAY_URDF = os.path.join(pd.getDataPath(), "tray/tray.urdf")
OBJECT_URDF = os.path.join(pd.getDataPath(), "duck_vhacd.urdf")


class PickAndPlaceEnvironment(TemplateEnvironment):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.text = grip.robot.add_text(
            [-0.5, 0, 1.0],
            text="Py-Grip Package: Table Pick & Place Environment",
            cid=self.world.id,
        )

        self.world.register_keyhandler(ord("r"), self.reset)

    def init(self):
        super().init()

        # return
        # spawn reminder of the objects
        self.random_offset = np.random.randn(3) * 0
        tray_pos = [
            0.3 + self.random_offset[0],
            -0.7 + self.random_offset[1],
            0.0,
        ]
        tray_ori = [0.0, 0.0, 0.0, 1.0]
        self.tray = g.BulletObject(
            world=self.world,
            urdf_file=TRAY_URDF,
            position=tray_pos,
            orientation=tray_ori,
            fixed=True,
            force_concave=True,
            compliant_dyn=True,
        )

        object_rel_pos = [0.0, 0.0, 0.1]
        object_rel_ori = [0.0, 0.0, 0.0, 1.0]

        object_pos, object_ori = grip.math.multiply_transform(
            tray_pos, tray_ori, object_rel_pos, object_rel_ori
        )

        self.object = g.BulletObject(
            world=self.world,
            urdf_file=OBJECT_URDF,
            position=object_pos,
            orientation=object_ori,
            fixed=False,
            compliant_dyn=True,
        )

        self.add("tray", self.tray)
        self.add("object", self.object)


if __name__ == "__main__":
    env = PickAndPlaceEnvironment()

    while p.isConnected():
        env.step(None)



import abc


class Reward:
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def reward(self, env, *args, **kwargs):
        """
        Computes reward based on arbitrary list of parameters (application/task dependent)
        """
        raise NotImplementedError



import logging

LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
    "not_set": logging.NOTSET,
}

log = None


class Logger:
    def __init__(self, name, level_name="debug"):
        self.configure(name, level_name)

    def configure(self, name, level_name="debug"):
        format_str = "[%(asctime)s][%(levelname)s]: %(message)s"

        level = LEVELS.get(level_name, logging.NOTSET)
        logging.basicConfig(format=format_str, level=level)

        self._logger = logging.getLogger(name)
        self._logger.propagate = False
        # otherwise handler additions are propagated to root logger resulting in double printing

        # creating console handler
        self._ch = logging.StreamHandler()
        self._ch.setLevel(level)

        self._formatter = logging.Formatter(format_str)
        self._ch.setFormatter(self._formatter)

        # add handler only if it doesn't already have one
        if not self._logger.handlers:
            self._logger.addHandler(self._ch)

        self._logger.setLevel(level)

    def debug(self, msg, *args, **kwargs):
        self._logger.debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self._logger.info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self._logger.warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self._logger.error(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        self._logger.critical(msg, *args, **kwargs)

    def setLevel(self, level_name):
        level = LEVELS.get(level_name, logging.NOTSET)

        self._ch.setLevel(level)
        self._logger.setLevel(level)


def logging_setup(name=None, level_name="debug"):
    global log
    log = Logger(name, level_name)
    return log


log = logging_setup("rosout", "info")


def set_log_level(level_name):
    level = LEVELS.get(level_name, logging.NOTSET)

    l = logging.getLogger("rosout")
    l.setLevel(level=level)



import pybullet as p
import numpy as np
from typing import Union, List, Tuple


def invert_transformation_matrix(transform_matrix: np.ndarray):
    """
    Inverts a 4x4 transformation matrix without using matrix inverse.
    Args: T (np.ndarray: shape-(4x4), a valid transformation matrix)

    """
    inverse_rot = transform_matrix[:3, :3].T
    inverse_pos = -inverse_rot @ transform_matrix[:3, 3]

    inverse = np.eye(4)
    inverse[:3, :3] = inverse_rot
    inverse[:3, 3] = inverse_pos

    return inverse


def invert_transform(
    position: Union[np.ndarray, List[float]], quaternion: Union[np.ndarray, List[float]]
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Returns the inverse of the given transform

    Args:
        position (np.ndarray): array-like shape-(3,) representing a position/translation
        quaternion (np.ndarray): array-like shape-(4,) unit quaternion
    Returns:
        (tuple[np.ndarray,np.ndarray]): position and orientation representing the inverse of the given transform
    """
    pos, quat = p.invertTransform(position, quaternion)
    return np.asarray(pos), np.asarray(quat)


def multiply_transform(
    pA: Union[np.ndarray, List[float]],
    qA: Union[np.ndarray, List[float]],
    pB: Union[np.ndarray, List[float]],
    qB: Union[np.ndarray, List[float]],
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Multiply two transforms given by their respective positions and orientation quaternions

    Args:
        pA (np.ndarray): array-like shape-(3,) representing a position/translation
        qA (np.ndarray): array-like shape-(4,) unit quaternion
        pB (np.ndarray): array-like shape-(3,) representing a position/translation
        qB (np.ndarray): array-like shape-(4,) unit quaternion
    Returns:
        (tuple[np.ndarray,np.ndarray]): position and orientation representing the multiplied transforms
    """
    pos, quat = p.multiplyTransforms(pA, qA, pB, qB)
    return np.asarray(pos), np.asarray(quat)


def transform_inv(points: np.ndarray, transform: np.ndarray) -> np.ndarray:
    """
    Transform points using the inverse of the given transform

    Args:
        points (np.ndarray): array-like shape-(M,3) representing a list of 3-dimensional points
        transform (np.ndarray): array-like shape-(4,4) transformation matrix
    Returns:
        (np.ndarray): array-like shape-(M,3) representing a list of 3-dimensional transformed points
    """
    R = transform[:3, :3].T
    t = transform[:3, 3]

    local_points = np.dot(R, (points - t).T).T

    return local_points


def transform(points, transform):
    """
    Transform points given a rigid body transformation matrix

    Args:
        points (np.ndarray): array-like shape-(M,3) representing a list of 3-dimensional points
        transform (np.ndarray): array-like shape-(4,4) transformation matrix
    Returns:
        (np.ndarray): array-like shape-(M,3) representing a list of 3-dimensional transformed points
    """
    R = transform[:3, :3]
    t = transform[:3, 3]

    transformed_points = np.dot(R, points.T).T + t.reshape(1, 3)

    return transformed_points



import numpy as np
from typing import Union, List
from scipy.spatial.transform import Rotation as R
from scipy.spatial.transform import Slerp

from transforms3d.euler import quat2euler

# axis sequences for Euler angles
_NEXT_AXIS = [1, 2, 0, 1]

# map axes strings to/from tuples of inner axis, parity, repetition, frame
_AXES2TUPLE = {
    "sxyz": (0, 0, 0, 0),
    "sxyx": (0, 0, 1, 0),
    "sxzy": (0, 1, 0, 0),
    "sxzx": (0, 1, 1, 0),
    "syzx": (1, 0, 0, 0),
    "syzy": (1, 0, 1, 0),
    "syxz": (1, 1, 0, 0),
    "syxy": (1, 1, 1, 0),
    "szxy": (2, 0, 0, 0),
    "szxz": (2, 0, 1, 0),
    "szyx": (2, 1, 0, 0),
    "szyz": (2, 1, 1, 0),
    "rzyx": (0, 0, 0, 1),
    "rxyx": (0, 0, 1, 1),
    "ryzx": (0, 1, 0, 1),
    "rxzx": (0, 1, 1, 1),
    "rxzy": (1, 0, 0, 1),
    "ryzy": (1, 0, 1, 1),
    "rzxy": (1, 1, 0, 1),
    "ryxy": (1, 1, 1, 1),
    "ryxz": (2, 0, 0, 1),
    "rzxz": (2, 0, 1, 1),
    "rxyz": (2, 1, 0, 1),
    "rzyz": (2, 1, 1, 1),
}

_TUPLE2AXES = dict((v, k) for k, v in _AXES2TUPLE.items())


def position_from_matrix(matrix):
    """
    Returns the translation part of a 4x4 homegeneous transform matrix

    Args:
        matrix (np.ndarray): A shape-(4,4) matrix

    Returns:
        (np.ndarray): A shape-(3,) translation part of the matrix
    """
    return matrix[:3, 3]


def quaternion_from_matrix(matrix: np.ndarray, isprecise=True) -> np.ndarray:
    """Return quaternion from rotation matrix.

    returned quaternion format (JPL): [x, y, z, w]

    If isprecise is True, the input matrix is assumed to be a precise rotation
    matrix and a faster algorithm is used.

    >>> q = quaternion_from_matrix(numpy.identity(4), True)
    >>> numpy.allclose(q, [1, 0, 0, 0])
    True
    >>> q = quaternion_from_matrix(numpy.diag([1, -1, -1, 1]))
    >>> numpy.allclose(q, [0, 1, 0, 0]) or numpy.allclose(q, [0, -1, 0, 0])
    True
    >>> R = rotation_matrix(0.123, (1, 2, 3))
    >>> q = quaternion_from_matrix(R, True)
    >>> numpy.allclose(q, [0.9981095, 0.0164262, 0.0328524, 0.0492786])
    True
    >>> R = [[-0.545, 0.797, 0.260, 0], [0.733, 0.603, -0.313, 0],
    ...      [-0.407, 0.021, -0.913, 0], [0, 0, 0, 1]]
    >>> q = quaternion_from_matrix(R)
    >>> numpy.allclose(q, [0.19069, 0.43736, 0.87485, -0.083611])
    True
    >>> R = [[0.395, 0.362, 0.843, 0], [-0.626, 0.796, -0.056, 0],
    ...      [-0.677, -0.498, 0.529, 0], [0, 0, 0, 1]]
    >>> q = quaternion_from_matrix(R)
    >>> numpy.allclose(q, [0.82336615, -0.13610694, 0.46344705, -0.29792603])
    True
    >>> R = random_rotation_matrix()
    >>> q = quaternion_from_matrix(R)
    >>> is_same_transform(R, quaternion_matrix(q))
    True
    >>> is_same_quaternion(quaternion_from_matrix(R, isprecise=False),
    ...                    quaternion_from_matrix(R, isprecise=True))
    True
    >>> R = euler_matrix(0.0, 0.0, numpy.pi/2.0)
    >>> is_same_quaternion(quaternion_from_matrix(R, isprecise=False),
    ...                    quaternion_from_matrix(R, isprecise=True))
    True

    """
    M = np.array(matrix, dtype=np.float64, copy=False)[:4, :4]
    if isprecise:
        q = np.empty((4,))
        t = np.trace(M)
        if t > M[3, 3]:
            q[0] = t
            q[3] = M[1, 0] - M[0, 1]
            q[2] = M[0, 2] - M[2, 0]
            q[1] = M[2, 1] - M[1, 2]
        else:
            i, j, k = 0, 1, 2
            if M[1, 1] > M[0, 0]:
                i, j, k = 1, 2, 0
            if M[2, 2] > M[i, i]:
                i, j, k = 2, 0, 1
            t = M[i, i] - (M[j, j] + M[k, k]) + M[3, 3]
            q[i] = t
            q[j] = M[i, j] + M[j, i]
            q[k] = M[k, i] + M[i, k]
            q[3] = M[k, j] - M[j, k]
            q = q[[3, 0, 1, 2]]
        q *= 0.5 / np.sqrt(t * M[3, 3])
    else:
        m00 = M[0, 0]
        m01 = M[0, 1]
        m02 = M[0, 2]
        m10 = M[1, 0]
        m11 = M[1, 1]
        m12 = M[1, 2]
        m20 = M[2, 0]
        m21 = M[2, 1]
        m22 = M[2, 2]
        # symmetric matrix K
        K = np.array(
            [
                [m00 - m11 - m22, 0.0, 0.0, 0.0],
                [m01 + m10, m11 - m00 - m22, 0.0, 0.0],
                [m02 + m20, m12 + m21, m22 - m00 - m11, 0.0],
                [m21 - m12, m02 - m20, m10 - m01, m00 + m11 + m22],
            ]
        )
        K /= 3.0
        # quaternion is eigenvector of K that corresponds to largest eigenvalue
        w, V = np.linalg.eigh(K)
        q = V[[3, 0, 1, 2], np.argmax(w)]
    if q[0] < 0.0:
        np.negative(q, q)

    # Converting to [x, y, z, w] (JPL format)
    q = np.array([q[1], q[2], q[3], q[0]])
    q /= np.linalg.norm(q)
    return q


def quaternion_from_matrix2(R: np.ndarray) -> np.ndarray:
    """
    Converts a rotation matrix to an orientation quaternion.

    Args:
        R (numpy.ndarray): A shape-(4, 4) or shape-(3, 3) array
    Returns:
        (numpy.ndarray): A shape-(4,) array representing a unity quaternion
    """
    tc = np.trace(R[:3, :3])

    q = np.zeros((3, 1))
    if tc > 0:
        s = 2 * np.sqrt(1 + tc)
        qw = 0.25 * s  # s = 4 * qw

        q = np.array(
            [
                (R[2][1] - R[1][2]) / s,
                (R[0][2] - R[2][0]) / s,
                (R[1][0] - R[0][1]) / s,
                qw,
            ]
        )

    elif R[0, 0] > R[1, 1] and R[0, 0] > R[2, 2]:
        s = np.sqrt(1.0 + R[0, 0] - R[1, 1] - R[2, 2]) * 2  # s = 4 * qx
        qw = (R[2, 1] - R[1, 2]) / s
        qx = 0.25 * s
        qy = (R[0, 1] + R[1, 0]) / s
        qz = (R[0, 2] + R[2, 0]) / s

        q = np.array([qx, qy, qz, qw])
    elif R[1, 1] > R[2, 2]:
        s = np.sqrt(1.0 + R[1, 1] - R[0, 0] - R[2, 2]) * 2  # S = 4 * qy
        qw = (R[0, 2] - R[2, 0]) / s
        qx = (R[0, 1] + R[1, 0]) / s
        qy = 0.25 * s
        qz = (R[1, 2] + R[2, 1]) / s

        q = np.array([qx, qy, qz, qw])
    else:
        s = np.sqrt(1.0 + R[2, 2] - R[0, 0] - R[1, 1]) * 2  # // S = 4 * qz
        qw = (R[1, 0] - R[0, 1]) / s
        qx = (R[0, 2] + R[2, 0]) / s
        qy = (R[1, 2] + R[2, 1]) / s
        qz = 0.25 * s

        q = np.array([qx, qy, qz, qw])

    return q


def quaternion2rotation_matrix(
    quaternion: Union[np.ndarray, List[float]]
) -> np.ndarray:
    """
    Returns a rotation matrix from a quaternion

    Args:
        quaternion (np.ndarray|list[float]): unit quaternion 1 quaternion: [x,y,z,w] (shape used in Bullet Physics) to rotation matrix
            see: http://www.mrelusive.com/publications/papers/SIMD-From-Quaternion-to-Matrix-and-Back.pdf

    Returns:
        (np.ndarray): A shape-(4,4) array representing a rotation matrix
    """

    if isinstance(quaternion, list):
        quaternion = np.array(quaternion, dtype=np.float64)

    n = np.dot(quaternion, quaternion)
    if n < 1e-12:
        return np.identity(3)
    q = np.array([quaternion[3], quaternion[0], quaternion[1], quaternion[2]])
    q *= np.sqrt(2.0 / n)
    q = np.outer(q, q)
    return np.array(
        [
            [1.0 - q[2, 2] - q[3, 3], q[1, 2] - q[3, 0], q[1, 3] + q[2, 0]],
            [q[1, 2] + q[3, 0], 1.0 - q[1, 1] - q[3, 3], q[2, 3] - q[1, 0]],
            [q[1, 3] - q[2, 0], q[2, 3] + q[1, 0], 1.0 - q[1, 1] - q[2, 2]],
        ]
    )


def position_quaternion2matrix(p: np.ndarray, q: np.ndarray) -> np.ndarray:
    """
    Converts a position and orientation quaternion to a 4x4 transformation matrix

    Args:
        p (numpy.ndarray): A shape-(3,) array
        q (numpy.ndarray): A shape-(4,) array unit quaternion
    Returns:
        (numpy.ndarray): A shape-(4,4) array representing a rigid body transform
    """
    transform = np.eye(4)

    R = quaternion2rotation_matrix(q)

    transform[0:3, 0:3] = R
    transform[:3, 3] = p

    return transform


def matrix2position_quaternion(matrix: np.ndarray) -> np.ndarray:
    """
    Converts a rotation matrix to an orientation quaternion.

    Args:
        matrix (numpy.ndarray): A shape-(4, 4) array
    Returns:
        (numpy.ndarray): A shape-(4,) array representing a unity quaternion
    """

    p = position_from_matrix(matrix)
    q = quaternion_from_matrix(matrix)

    return p, q


def make_slerp(key_times: np.ndarray, key_quaternions: List[np.ndarray]) -> Slerp:
    """
    Creates a spherical linear interpolation object

    Args:
        key_times (numpy.ndarray): A shape-(N,) array
        key_quaternions (list[np.ndarray]): A list of orientation quaternions
    Returns:
        (scipy.spatial.transform.Slerp): A slerp object
    """

    rotations = R.from_quat(key_quaternions)

    slerp = Slerp(key_times, rotations)

    return slerp


def quaternion_mult(
    q1: Union[np.ndarray, List[float]], q2: Union[np.ndarray, List[float]]
) -> np.ndarray:
    """
    Quaternion multiplication convention: same order as matrix multiplication
    so Ra*Rb is same as qa*qb
    See "Why and How to Avoid the Flipped Quaternion Multiplication"
    paper: https://arxiv.org/abs/1801.07478
    comment: same convention as ROS

    Args:
        q1 (np.ndarray|list[float]): unit quaternion 1
        q2 (np.ndarray|list[float]): unit quaternion 2
    Returns:
        (np.ndarray): quaternion multiplication q1*q2
    """

    if isinstance(q1, list):
        q1 = np.array(q1)
    if isinstance(q2, list):
        q2 = np.array(q2)

    s1 = q1[3]
    s2 = q2[3]
    v1 = q1[0:3]
    v2 = q2[0:3]

    dot = np.multiply(s1, s2) - np.dot(v1, v2)
    mult1 = np.multiply(s1, v2)
    mult2 = np.multiply(s2, v1)
    result = np.hstack([mult1 + mult2 + np.cross(v1, v2), dot])

    return result


def quat_multiply(
    q1: Union[np.ndarray, List[float]], q0: Union[np.ndarray, List[float]]
) -> np.ndarray:
    """Return multiplication of two quaternions.

    Args:
        q1 (np.ndarray|list[float]): array-like shape-(4,) unit quaternion 1
        q2 (np.ndarray|list[float]): array-like shape-(4,) unit quaternion 2
    Returns:
        (np.ndarray): shape-(4,) quaternion multiplication q1*q2
    Examples:
        >>> q = quat_multiply([1, -2, 3, 4], [-5, 6, 7, 8])
        >>> numpy.allclose(q, [-44, -14, 48, 28])
        True
    """
    x0, y0, z0, w0 = q0
    x1, y1, z1, w1 = q1
    return np.array(
        (
            x1 * w0 + y1 * z0 - z1 * y0 + w1 * x0,
            -x1 * z0 + y1 * w0 + z1 * x0 + w1 * y0,
            x1 * y0 - y1 * x0 + z1 * w0 + w1 * z0,
            -x1 * x0 - y1 * y0 - z1 * z0 + w1 * w0,
        ),
        dtype=np.float64,
    )


def quaternion_diff(
    q1: Union[np.ndarray, List[float]], q2: Union[np.ndarray, List[float]]
) -> np.ndarray:
    """Return the difference between two quaternions.
    The error can be used for cartesian feedback controllers and has been originally proposed by:
    Yuan88: https://www.cs.cmu.edu/~cga/dynopt/readings/Yuan88-quatfeedback.pdf

    Args:
        q1 (np.ndarray|list[float]): array-like shape-(4,) unit quaternion 1
        q2 (np.ndarray|list[float]): array-like shape-(4,) unit quaternion 2
    Returns:
        (np.ndarray): shape-(4,) q1*quaternion_conj(q2)
    """
    return quaternion_mult(q1, quaternion_conj(q2))


def quaternion_error(
    q1: Union[np.ndarray, List[float]], q2: Union[np.ndarray, List[float]]
) -> float:
    """Return the quaternion error, which is the magnitude of the quaternion difference

    Args:
        q1 (np.ndarray|list[float]): array-like shape-(4,) unit quaternion 1
        q2 (np.ndarray|list[float]): array-like shape-(4,) unit quaternion 2
    Returns:
        (float): quaternion error

    """
    qdiff = quaternion_diff(q1, q2)

    return np.linalg.norm(qdiff[:3])


def compute_log(q: np.ndarray) -> np.ndarray:
    """Return the logarithmic map of the quaternion

    Args:
        q (np.ndarray): array-like shape-(4,) unit quaternion 1
    Returns:
        (np.ndarray): shape-(3,) array representing the log map of the unit quaternion

    """
    v = q[:3]
    norm_q = np.linalg.norm(q)
    norm_v = np.linalg.norm(v)

    tolerance = 1e-6

    if norm_v < tolerance:
        return np.zeros(3)
    else:
        return 2 * np.arccos(q[3] / norm_q) * v / norm_v


def exp_map(omg: np.ndarray) -> np.ndarray:
    """Return the exponential map of 3d vector

    Args:
        q (np.ndarray): array-like shape-(3,) vector
    Returns:
        (np.ndarray): a unit quaternion shape-(4,) array representing the exp map of the vector

    """

    angle = np.linalg.norm(omg)
    if angle == 0:
        return np.asarray([0, 0, 0, 1])
    return axis_angle2quaternion(omg / angle, angle)


def compute_omg(q1: np.ndarray, q2: np.ndarray) -> np.ndarray:
    """Return the angular difference between two quaternions

    Args:
        q1 (np.ndarray): array-like shape-(4,) unit quaternion 1
        q2 (np.ndarray): array-like shape-(4,) unit quaternion 2
    Returns:
        (np.ndarray): A shape-(3,) array representing 3D angular difference (radians)

    """
    dot = np.dot(q1, q2)

    if dot < 0:
        q2 = -q2

    return 2.0 * compute_log(quaternion_mult(q1, quaternion_conj(q2)))


def rpy2quaternion(rpy: Union[np.ndarray, List[float]]) -> np.ndarray:
    """
    Return a quaternion from roll pitch yaw vector (radians)

    Args:
        rpy (np.ndarray): array-like shape-(3,) [roll, pitch, yaw] (radians)
    Returns:
        (np.ndarray): a unit quaternion shape-(4,)

    """
    return euler2quaternion(rpy[0], rpy[1], np.fmod(rpy[2], 2 * np.pi), axes="sxyz")


def quaternion2rpy(q: np.ndarray) -> np.ndarray:
    """
    Return a roll pitch yaw vector (radians) from a quaternion

    Args:
        q (np.ndarray): array-like shape-(4,) unit quaternion
    Returns:
        (np.ndarray): An array shape-(3,) containing the roll pitch and yaw angles (radians)

    """
    return quat2euler(q, axes="sxyz")


def euler2quaternion(ai, aj, ak, axes="sxyz"):
    """Return quaternion from Euler angles and axis sequence.
    ai, aj, ak : Euler's roll, pitch and yaw angles
    axes : One of 24 axis sequences as string or encoded tuple

    Examples:
        >>> q = _euler2quaternion(1, 2, 3, 'ryxz')
        >>> numpy.allclose(q, [0.310622, -0.718287, 0.444435, 0.435953])
        True
    """
    try:
        firstaxis, parity, repetition, frame = _AXES2TUPLE[axes.lower()]
    except (AttributeError, KeyError):
        _ = _TUPLE2AXES[axes]
        firstaxis, parity, repetition, frame = axes

    i = firstaxis
    j = _NEXT_AXIS[i + parity]
    k = _NEXT_AXIS[i - parity + 1]

    if frame:
        ai, ak = ak, ai
    if parity:
        aj = -aj

    ai /= 2.0
    aj /= 2.0
    ak /= 2.0
    ci = np.cos(ai)
    si = np.sin(ai)
    cj = np.cos(aj)
    sj = np.sin(aj)
    ck = np.cos(ak)
    sk = np.sin(ak)
    cc = ci * ck
    cs = ci * sk
    sc = si * ck
    ss = si * sk

    quaternion = np.empty((4,), dtype=np.float64)
    if repetition:
        quaternion[i] = cj * (cs + sc)
        quaternion[j] = sj * (cc + ss)
        quaternion[k] = sj * (cs - sc)
        quaternion[3] = cj * (cc - ss)
    else:
        quaternion[i] = cj * sc - sj * cs
        quaternion[j] = cj * ss + sj * cc
        quaternion[k] = cj * cs - sj * sc
        quaternion[3] = cj * cc + sj * ss
    if parity:
        quaternion[j] *= -1

    return quaternion


def quaternion_conj(q: np.ndarray) -> np.ndarray:
    """
    Return the conjugate of the given quaternion

    Args:
        q (np.ndarray): array-like shape-(4,) unit quaternion
    Returns:
        (np.ndarray): quaternion conjugate

    """
    return np.hstack([-q[0:3], q[3]])


def axis_angle2quaternion(axis: np.ndarray, angle: np.ndarray) -> np.ndarray:
    """
    Return a quaternion from an axis angle representation

    Args:
        axis (np.ndarray): array-like shape-(3,) unit vector
        angle (float): angle in radians
    Returns:
        (np.ndarray): array shape-(4,) unit quaternion

    """
    return np.hstack([np.array(axis) * np.sin(angle * 0.5), np.cos(angle * 0.5)])



#!/usr/bin/env python3

import time
import numpy as np

import grip
import os
import pybullet_data as pd


if __name__ == "__main__":
    world = grip.robot.BulletWorld(phys_opt="gui")

    plane = grip.robot.BulletObject(
        urdf_file=os.path.join(pd.getDataPath(), "plane.urdf"), position=[0, 0, 0]
    )

    duck = grip.robot.BulletObject(
        urdf_file=os.path.join(pd.getDataPath(), "duck_vhacd.urdf"),
        position=[0, 0, 0.5],
        scale=8.0,
        fixed=False,
    )

    camera = grip.sensors.RGBDCamera(
        cid=world.id,
        position=np.array([0, 0, 1.5]),
        orientation=grip.math.rpy2quaternion([0.0, np.pi, 0.0]),
    )

    camera.draw_frustum(0.0)

    while True:
        rgb, depth, seg = camera.obs()

        time.sleep(0.1)



import time
import warnings
import pickle
from os.path import exists, join, dirname, abspath
import os
import contextlib
import importlib

from .logging_utils import log

from functools import wraps

PROF_DATA = {}


def profile(fn):
    @wraps(fn)
    def with_profiling(*args, **kwargs):
        start_time = time.time()

        ret = fn(*args, **kwargs)

        elapsed_time = time.time() - start_time

        if fn.__name__ not in PROF_DATA:
            PROF_DATA[fn.__name__] = [0, []]
        PROF_DATA[fn.__name__][0] += 1
        PROF_DATA[fn.__name__][1].append(elapsed_time)

        return ret

    return with_profiling


def print_prof_data():
    for fname, data in PROF_DATA.items():
        max_time = max(data[1])
        avg_time = sum(data[1]) / len(data[1])
        print("Function %s called %d times. " % (fname, data[0]))
        print("Execution time max: %.3f, average: %.3f" % (max_time, avg_time))


def clear_prof_data():
    global PROF_DATA
    PROF_DATA = {}


def get_package_path(package_name="grip_assets", path_suffix=""):
    m = None
    try:
        m = importlib.import_module(package_name)
    except ModuleNotFoundError as exec:
        log.info("Module {} was not found.".format(package_name))
        if package_name == "grip_data":
            log.info(
                "The package grip_data can be downloaded, would you like to do that now?"
            )
            answer = input("(y/n)? ")
            if answer in ["y", "Y"]:
                gs = importlib.import_module("grip_assets")
                gs.fetch_data()

                log.info("Package grip_data was downloaded and installed.")

                m = importlib.import_module(package_name)
        else:
            raise ModuleNotFoundError from exec

    package_path = dirname(m.__file__)

    package_path = join(package_path, path_suffix)

    return package_path


def get_abs_path(path):
    return abspath(path)


def crawl(path, extension_filter="urdf"):
    crawled_paths = dict()

    repeat_count = 0
    for dirpath, _, files in os.walk(path):
        for file in files:
            prefix = ""
            key = file
            if crawled_paths.get(file, None) is not None:
                prefix = "{:04d}".format(repeat_count)
                key = "{}_{}".format(prefix, file)
                repeat_count += 1

            if extension_filter is None:
                crawled_paths[key] = os.path.join(dirpath, file)
            elif file.split(".")[-1] == extension_filter:
                crawled_paths[key] = os.path.join(dirpath, file)

    return crawled_paths


def file_list(path, extension_filter="urdf"):
    files = crawl(path, extension_filter=extension_filter)

    fs = [file_path for _, file_path in files.items()]
    fs.sort()

    return fs


def get_file_path(file, search_paths):
    if not isinstance(search_paths, list):
        search_paths = [search_paths]

    crawled_paths = {}
    ext = file.split(".")[-1]

    for path in search_paths:
        crawled_paths.update(crawl(path, ext))

    file_path = crawled_paths.get(file, None)

    if file_path is None:
        log.warning("File %s not found, returning None path." % (file,))

    return file_path


def save_data(data, filename, over_write_existing=False):
    if not over_write_existing:
        if exists(filename):
            warnings.warn("File exist by same name, renaming new file...")
            filename = (
                filename[:-4]
                + time.strftime("_%b_%d_%Y_%H_%M_%S", time.localtime())
                + ".pkl"
            )

    output = open(filename, "wb")

    # Pickle dictionary using protocol 0.
    pickle.dump(data, output)

    # Pickle the list using the highest protocol available.
    # pickle.dump(selfref_list, output, -1)

    output.close()


def load_data(filename):
    try:
        pkl_file = open(filename, "rb")
    except OSError as e:
        raise e

    data = pickle.load(pkl_file)

    pkl_file.close()

    return data


def import_module(module_name):
    lib = None
    try:
        lib = importlib.import_module(module_name)
    except:
        log.debug(f"Failed to import module: {module_name}. Has it been installed?")

    return lib


@contextlib.contextmanager
def pushd(new_dir):
    old_dir = os.getcwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(old_dir)


def read_file_as_str(filename):
    with open(filename, "r") as file:
        data_str = file.read().rstrip()

    return data_str



import rclpy
from sensor_msgs.msg import JointState
from typing import Callable, List
import numpy as np
from rclpy.callback_groups import ReentrantCallbackGroup, MutuallyExclusiveCallbackGroup


class JointStatePub:
    """A joint state publisher class responsible for publishing all joint states of a grip.robot.BulletRobot."""

    def __init__(
        self,
        parent_node: "rclpy.node.Node",
        robot: "grip.robot.BulletRobot",
        topic_name: str = "/joint_states",
        command_topic_name: str = "/joint_commands",
        **kwargs,
    ):
        """Constructs a JointStatePub object

        Args:
            parent_node (rclpy.node.Node): parent node to which this joint state publisher is associated with.
            robot (grip.robot.BulletRobot): robot from which joint states should be published.
            topic_name (str, optional): the desired topic name that the joint states should be published to. Defaults to "/joint_states".
        """

        self.lock = kwargs.get("lock", None)
        self.topic_name = topic_name
        self.robot = robot
        self.node = parent_node

        # register this node in the network as a publisher in /joint_states topic
        self.pub_joint_states = self.node.create_publisher(JointState, topic_name, 10)

        self.joint_command_sub = self.node.create_subscription(
            JointState,
            command_topic_name,
            self.joint_command_callback,
            1,
        )

        self.joint_msg = JointState()
        self.joint_msg.name = [name for name, _ in self.robot.joint_dict.items()]
        self._timer_func = None
        self._last_joint_command = None

    def execute(self) -> None:
        """Single step execution: reads current joint states and publishes it."""

        if self.lock is not None:
            with self.lock:
                self._execute()
        else:
            self._execute()

    def _execute(self) -> None:
        """Single step execution: reads current joint states and publishes it.

        This function gets called from pybullet ros main update loop
        """
        # setup msg placeholder
        joint_msg = JointState()
        # get joint states
        for joint_name, joint_info in self.robot.joint_dict.items():
            # get joint state from pybullet

            joint_state = self.robot.joint_angle(joint_info["idx"])
            # fill msg
            joint_msg.name.append(joint_name)
            joint_msg.position.append(joint_state[0])
            joint_msg.velocity.append(joint_state[1])
            joint_msg.effort.append(joint_state[2])  # applied effort in last sim step

        # update msg time using ROS time api
        joint_msg.header.stamp = self.node.get_clock().now().to_msg()
        # publish joint states to ROS
        self.pub_joint_states.publish(joint_msg)

        self.command_update()

    def _publisher_callback(self) -> None:
        """Publisher callback, used for asynchronous publishing using timers."""
        try:
            self.execute()
        except Exception as e:
            self.node.get_logger().warning(
                f"Async joint state publisher failed to publish joint state. Message {e}",
                once=True,
            )

    def setup_timer_publisher(self, rate: int = 10) -> None:
        """Sets up a timer function that periodically publishes joint states at specified rate.

        Args:
            rate (int, optional): desired publishing rate. Defaults to 10.
        """
        self._timer_cb_group = MutuallyExclusiveCallbackGroup()
        self._timer_func = self.node.create_timer(
            1.0 / rate, self._publisher_callback, callback_group=self._timer_cb_group
        )

    def stop_timer_publisher(self) -> None:
        """Stops the timer function"""

        if self._timer_func is None:
            return

        self._timer_func.destroy()
        self._timer_func = None

    def joint_command_callback(self, msg: JointState) -> None:
        """Callback for the joint command subscriber."""
        # print("Received command: ", msg)
        self._last_joint_command = msg

    def command_update(self):
        if self._last_joint_command is not None and np.any(
            self._last_joint_command.position
        ):
            self.robot.set_angles(
                self._last_joint_command.position,
                joint_names=self._last_joint_command.name,
            )

    def destroy(self) -> None:
        """Destroy the underlying ROS handles."""

        self.stop_timer_publisher()
        self.pub_joint_states.destroy()
        self.pub_joint_states = None
        self.robot = None
        self.node = None


class LocalJointStatePub(JointStatePub):
    """Publishes subset of joint states as reported by robot.BulletRobot joint_names and get_joint_states"""

    def __init__(
        self,
        parent_node: "rclpy.node.Node",
        robot: "grip.robot.BulletRobot",
        topic_name: str = "/gripper/joint_states",
        command_topic_name: str = "/gripper/joint_commands",
        **kwargs,
    ):
        super().__init__(parent_node, robot, topic_name, command_topic_name, **kwargs)

        self.joint_msg = JointState()
        self.joint_msg.name = self.robot.joint_names
        self._timer_func = None
        self._last_joint_command = None

    def _execute(self) -> None:
        """Single step execution: reads current joint states and publishes it.

        This function gets called from pybullet ros main update loop
        """
        # setup msg placeholder
        joint_msg = JointState()

        joint_names = self.robot.joint_names
        position, velocity, _, effort = self.robot.get_joint_state()

        joint_msg.name = joint_names
        joint_msg.position = position.tolist()
        joint_msg.velocity = velocity.tolist()
        joint_msg.effort = effort.tolist()

        # update msg time using ROS time api
        joint_msg.header.stamp = self.node.get_clock().now().to_msg()
        # publish joint states to ROS
        self.pub_joint_states.publish(joint_msg)

        self.command_update()


class JointStateSub:
    """A joint state subscriber class responsible for subscribing to the joint state of an external robot and mirroring this state to a grip.robot.BulletRobot instance."""

    def __init__(
        self,
        parent_node: "rclpy.node.Node",
        topic_name: str = "/joint_states",
        joint_update_callback: Callable[
            [np.ndarray, List[str]], None
        ] = lambda _, __: None,
        **kwargs,
    ):
        """Constructs a JointStatePub object

        Args:
            parent_node (rclpy.node.Node): parent node to which this joint state subscriber is associated with.
            topic_name (str, optional): the desired joint state topic name to subscribe to. Defaults to "/joint_states".
            joint_update_callback (Callable[[np.ndarray, List[str]], None]): callback used for updating joint states. Defaults to empty lambda.
        """

        self.lock = kwargs.get("lock", None)
        self.topic_name = topic_name
        self.node = parent_node

        # register this node in the network as a publisher in /joint_states topic
        self.sub_joint_states = self.node.create_subscription(
            JointState, self.topic_name, self.on_joint_state, 1
        )

        self._positions = None
        self._joint_names = None
        self._joint_update_callback = joint_update_callback

    def _register_joint_update_callback(
        self, joint_update_callback: Callable[[np.ndarray, List[str]], None]
    ) -> None:
        self._joint_update_callback = joint_update_callback

    def on_joint_state(self, joint_state: JointState) -> None:
        """Single step execution: receives current joint states and mirroes it to simulated robot.

        Args:

            joint_state: incoming joint state.
        """

        self._positions = np.array(joint_state.position)
        self._joint_names = joint_state.name

        # Sets simulated robot to same joint states received from external robot
        self._joint_update_callback(self._positions, self._joint_names)

    @property
    def positions(self) -> np.ndarray:
        """
        Current joint positions

        Returns:
            (np.ndarray): most recently received joint angles
        """

        return self._positions

    @property
    def joint_names(self) -> List[str]:
        """
        Get received joint names

        Returns:
            (List[str]): most recently received joint names
        """

        return self._joint_names

    def destroy(self):
        """Destroy the underlying ROS handles."""
        self.sub_joint_states.destroy()
        self.sub_joint_states = None



try:
    from pybullet_planning import (
        plan_joint_motion,
        get_custom_limits,
    )
    from pybullet_planning.utils import set_client
    from pybullet_planning.interfaces.env_manager.pose_transformation import all_between
except Exception as e:
    print(f"Failed to import pybullet_planning: {e}")

import numpy as np
from .path import Path


class PBBasicPlanner:
    """
    This is class supporting basic motion planning in cartesian and configuration space
    with obstacle avoidance on top of the pybullet_planning package, which is a suite of
    utility functions for robot motion planning, manipulation planning and task and
    motion planning package tailored for PyBullet simulation.


    Reference:
        Caelan Reed Garrett. PyBullet Planning. https://pypi.org/project/pybullet-planning/. 2018.

    Args:
        robot (grip.robot.BulletRobot): An instance of a bullet robot
    """

    def __init__(self, robot):
        self.robot = robot

    def format_joint_limits(self):
        """
        Parsing the joint limits for all joint of the target robot
        Slightly postprocessing it by adding a bit buffer to cope with reading error
        from PyBullet (Less optimal solution)

        Returns:
            custom_limits: [dict]
                        dictionary with key as joint name, value as tuple (lower_limit, upper_limit)
            lower_limits: [list]
                        list of lower limit for all joints in order
            upper_limits: [list]
                        list of upper limit for all joints in order
        """
        CLIENT = self.robot.phys_id
        set_client(CLIENT)

        lower_limits, upper_limits = get_custom_limits(
            self.robot.id, self.robot.joint_ids
        )
        # slightly enlarge the limits for pybullet reading error
        # for example0.04-->0.04000000xx
        custom_limits = {}
        lower_limits = list(lower_limits)
        upper_limits = list(upper_limits)
        for i, (joint, ll, ul) in enumerate(
            zip(self.robot.joint_ids, lower_limits, upper_limits)
        ):
            lower_limits[i] = ll - 0.01
            upper_limits[i] = ul + 0.01
            custom_limits[joint] = lower_limits[i], upper_limits[i]
        return custom_limits, lower_limits, upper_limits

    def plan_cartesian_goal(
        self,
        target_pose,
        obstacles=[],
        disabled_collisions=[],
        initial_conf=None,
        **kwargs,
    ):
        """
        Planing with BiRRT planner from initial configuration to final cartesian
        pose. The IK solver used here is from PyBullet which is aversion of Samuel
        Buss Inverse Kinematics library which also supports Null space control if
        lowerLimits, upperLimits, jointRanges, restPoses are specified.

        Args:
            target_pose: [list]
                    [[postion as list of 3 floats], [quaternion as list of 4 floats]
            obstacles: [list]
                    list of object indexes for target objects
            disabled_collisions: [list]
                    list of object indexes for object which temperoarily disable
                    collision checking, for example during grasping, the collision
                    checking on object or surface will be disabled
            initial_conf: [list]
                    joint angles for initial configuration in order. If None will compute
                    from the current configuration.

        Returns:
            path: [list]
                A trajectory from initial config to target pose as a list
                of joint configurations
        """
        # TODO: add disabled_collisions feature?
        CLIENT = self.robot.phys_id
        set_client(CLIENT)

        full_ret = kwargs.get("full_ret", False)
        kwargs.get("total_time", 5)
        dt = kwargs.get("dt", 0.001)

        original_conf = self.robot.angles
        if initial_conf is None:
            initial_conf = original_conf

        self.robot.reset_angles(initial_conf)
        # set_joint_positions(self.robot.id,
        #                    self.robot.joint_ids,
        #                    initial_conf)

        custom_limits, lower_limits, upper_limits = self.format_joint_limits()

        target_conf = self.robot.inverse_kinematics(*target_pose)

        print("Target conf: ", target_conf)
        path = plan_joint_motion(
            self.robot.id,
            self.robot.joint_ids,
            target_conf,
            obstacles=obstacles,
            self_collisions=False,
            custom_limits=custom_limits,
            max_distance=1e-3,
            iterations=1000,
            restarts=100,
            diagnosis=False,
        )

        ## Needs to reset robot state to where the were since pybullet_planner sets the joints of self.robot during planning
        self.robot.reset_angles(original_conf)

        # motion planner
        if (
            path is None
            or any(map(np.isnan, target_conf))
            or (not all_between(lower_limits, target_conf, upper_limits))
        ):
            target_conf = None
            return None
        else:
            if full_ret:
                gpath = Path(points=np.array(path), joint_names=self.robot.joint_names)

                return gpath
            else:
                return path

    def plan_joint_goal(self, target_conf, obstacles=[], initial_conf=None, **kwargs):
        """
        Planing with BiRRT planner from initial configuration to final configuration
        in joint space

        Args:
            target_pose: [list]
                    [[postion as list of 3 floats], [quaternion as list of 4 floats]
            obstacles: [list]
                    list of object indexes for target objects
            initial_conf: [list]
                    joint angles for initial configuration in order. If None will compute
                    from the current configuration.

        Returns:
            path: [list]
                A trajectory from initial config to target pose as a list
                of joint configurations
        """
        CLIENT = self.robot.phys_id
        set_client(CLIENT)

        full_ret = kwargs.get("full_ret", False)
        kwargs.get("total_time", 5)
        dt = kwargs.get("dt", 0.001)

        original_conf = self.robot.angles
        if initial_conf is None:
            initial_conf = original_conf

        self.robot.reset_angles(initial_conf)
        # set_joint_positions(self.robot.id,
        #                    self.robot.joint_ids,
        #                    initial_conf)

        custom_limits, lower_limits, upper_limits = self.format_joint_limits()

        path = plan_joint_motion(
            self.robot.id,
            self.robot.joint_ids,
            target_conf,
            obstacles=obstacles,
            self_collisions=False,
            custom_limits=custom_limits,
            max_distance=1e-2,
            iterations=1000,
            restarts=100,
            diagnosis=False,
        )

        ## Needs to reset robot state to where the were since pybullet_planner sets the joints of self.robot during planning
        self.robot.reset_angles(original_conf)

        # motion planner
        if any(map(np.isnan, target_conf)) or (
            not all_between(lower_limits, target_conf, upper_limits)
        ):
            target_conf = None
            return None
        else:
            if full_ret:
                gpath = Path(points=np.array(path), joint_names=self.robot.joint_names)

                error = np.linalg.norm(initial_conf - target_conf)
                if error > 1e-3:
                    return gpath.retime(self.robot, dt)
                else:
                    return gpath
            else:
                return path



from .base_env import *
from .reward import *
from .domain_randomiser import *
from .template_env import *
from .pick_and_place_env import *

__all__ = [
    "Env",
    "RobotEnv",
    "LightingRandomiser",
    "TextureRandomiser",
    "TemplateEnvironment",
    "PickAndPlaceEnvironment",
    "Reward",
]



#!/usr/bin/env python3

import time
import grip.robot as robot
from grip.robot import EndEffectorRegistry
from grip.ros import xacro_to_urdf_path

if __name__ == "__main__":
    world = robot.BulletWorld(phys_opt="gui")

    urdf_path = xacro_to_urdf_path(
        "robotiq_140_description",
        "robotiq_140_gripper_instance.urdf.xacro",
        "/tmp/robotiq_2f140.urdf",
        pkg_subfolder="urdf",
        xacro_args="hand:=true",
    )

    gripper = EndEffectorRegistry.make(
        "robotiq_2f140", urdf_file=urdf_path, phys_id=world.id, sliders=True
    )

    toggle_gripper = robot.BulletButton(world.id, "Toggle Gripper")
    is_closed = False
    while world.is_connected():
        if toggle_gripper.was_pressed():
            is_closed = not is_closed

            gripper.control_fingers("close" if is_closed else "open")

        time.sleep(0.1)



from .entity import Entity
from .robot_interface import RobotInterface
from .panda_gripper import *
from .world import *
from .bullet_robot import *
from .end_effector_registry import EndEffectorRegistry, EndEffectorType
from .panda_gripper import *
from .bullet_robotiq_2f_gripper import *
from .panda_robot import *
from .object import *
from .gui import *
from .types import *
from .articulated_object import ArticulatedObject
from .robot_arm import BulletRobotArm
from .collision import get_collision_info, is_colliding

__all__ = [
    "Entity",
    "RobotInterface",
    "GripperInterface",
    "BulletRobot",
    "RobotKinematics",
    "BulletObject",
    "BulletWorld",
    "BulletPandaArmGripper",
    "BulletPandaGripper",
    "BulletRobotiq2FGripper",
    "BulletRobotiq2F85Gripper",
    "BulletRobotiq2F140Gripper",
    "BulletSliders",
    "BulletButton",
    "addDebugFrame",
    "draw_frustum",
    "create_box",
    "EndEffectorRegistry",
    "EndEffectorType",
    "BulletDynamicsInfo",
    "BulletJointInfo",
    "ArticulatedObject",
    "BulletRobotArm",
    "load_sdf",
    "draw_cloud",
    "get_collision_info",
    "is_colliding",
]



# Copyright 2022 Sony Group Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rclpy.impl.implementation_singleton import rclpy_implementation as _rclpy
from rclpy.node import Node
from rclpy.signals import SignalHandlerGuardCondition
from rclpy.utilities import timeout_sec_to_nsec


def wait_for_message(msg_type, node: "Node", topic: str, time_to_wait=-1):
    """
    Wait for the next incoming message.
    :param msg_type: message type
    :param node: node to initialize the subscription on
    :param topic: topic name to wait for message
    :time_to_wait: seconds to wait before returning
    :return (True, msg) if a message was successfully received, (False, ()) if message
        could not be obtained or shutdown was triggered asynchronously on the context.
    """
    context = node.context
    wait_set = _rclpy.WaitSet(1, 1, 0, 0, 0, 0, context.handle)
    wait_set.clear_entities()

    sub = node.create_subscription(msg_type, topic, lambda _: None, 1)
    wait_set.add_subscription(sub.handle)
    sigint_gc = SignalHandlerGuardCondition(context=context)
    wait_set.add_guard_condition(sigint_gc.handle)

    timeout_nsec = timeout_sec_to_nsec(time_to_wait)
    wait_set.wait(timeout_nsec)

    subs_ready = wait_set.get_ready_entities("subscription")
    guards_ready = wait_set.get_ready_entities("guard_condition")

    if guards_ready:
        if sigint_gc.handle.pointer in guards_ready:
            return (False, None)

    if subs_ready:
        if sub.handle.pointer in subs_ready:
            msg_info = sub.handle.take_message(sub.msg_type, sub.raw)
            return (True, msg_info[0])

    return (False, None)



#!/usr/bin/env python3
import os
import numpy as np

try:
    import rclpy

except ImportError:
    raise RuntimeError("Unable to run example without ros2 installed.")

import pybullet_data as pd
import grip
from grip.ros import ROSRobotArm, ROSCamera

from rcl_interfaces.msg import ParameterDescriptor, ParameterType
from grip.ros.urdf_utils import normalise_urdf_paths


class ROSArmMoveitExample(rclpy.node.Node):
    """This node example will simulate a robot arm with a wrist camera."""

    def __init__(self, camera_fps: int = 10, joint_pub_rate: int = 150):
        super().__init__("robot_arm_moveit_example")

        self.declare_parameter(
            "robot_description",
            descriptor=ParameterDescriptor(type=ParameterType.PARAMETER_STRING),
        )
        self.declare_parameter(
            "enable_gui",
            descriptor=ParameterDescriptor(type=ParameterType.PARAMETER_BOOL),
        )

        # Creates bullet world with GUI
        self.world = grip.robot.BulletWorld(phys_opt="gui")

        # Creates ground plane object
        self.plane = grip.robot.BulletObject(
            urdf_file=os.path.join(pd.getDataPath(), "plane.urdf"), position=[0, 0, 0]
        )

        self.plane.set_colour([0.0, 0.7, 0.7, 1.0])
        self.duck = grip.robot.BulletObject(
            urdf_file=os.path.join(pd.getDataPath(), "duck_vhacd.urdf"),
            position=[0.7, 0.0, 0.0],
            scale=2.0,
            fixed=False,
        )

        urdf_path = normalise_urdf_paths(self.get_parameter("robot_description").value)

        # Creates robot arm
        self.arm = ROSRobotArm(
            parent_node=self,
            urdf_file=urdf_path,
            phys_id=self.world.id,
            ee_mount_link="panda_link8",
            tip_link="panda_hand",
            gripper_type="panda_hand",
            sliders=True,
            auto_sim=False,
            sim=True,
            state_topic="arm/joint_states",
            command_topic="arm/joint_commands",
        )

        print("Links: ", self.arm.get_link_names())

        # Creates ROSCamera object
        self.camera = ROSCamera(
            name="hand_camera",
            cid=self.world.id,
            position=np.array([0, 0, 1.5]),
            orientation=grip.math.rpy2quaternion([0.0, np.pi, 0.0]),
            parent_node=self,
            anchor_robot=self.arm,
            anchor_link="hand_camera_optical_frame",
            base_frame_id="panda_link0",
            sim=self.arm.sim,
        )

        self.send_goal_button = grip.robot.gui.BulletButton(
            self.world.id, "Execute Joint Goal"
        )

        if self.arm.sim:
            # If we are in sim, then we create timer publishers for camera and joint states
            # This is not needed if data is to be consumed synchrnously.
            # However, it is needed if one wants to simulate asynchronous publisers just like the real robot and cameras would be behave.

            # Start camera timer publisher
            self.camera.setup_timer_publisher(fps=camera_fps)
            self.arm.start_async_ros_comm(rate=joint_pub_rate)

            grip.robot.add_text([0, 0, 1], "mode: robot is simulated")

        else:
            grip.robot.add_text([0, 0, 1], "mode: robot is mirroring real robot state")

        self.timer = self.create_timer(1 / camera_fps, self.on_update)

    def on_update(self):
        if self.send_goal_button.was_pressed():
            gripper_goal = self.arm.gripper.joint_sliders.read()
            self.arm.gripper.set_angles(gripper_goal)


def main(args=None):
    rclpy.init(args=args)

    ros_arm_example = ROSArmMoveitExample()

    rclpy.spin(ros_arm_example)

    # Destroy the node explicitly
    # (optional - otherwise it will be done automatically
    # when the garbage collector destroys the node object)
    ros_arm_example.destroy_node()
    rclpy.shutdown()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import time
import grip.robot as robot
from grip.robot import EndEffectorRegistry


if __name__ == "__main__":
    world = robot.BulletWorld(phys_opt="gui")

    gripper = EndEffectorRegistry.make("panda_hand", phys_id=world.id, sliders=True)

    while world.is_connected():
        positions = gripper.joint_sliders.read()

        gripper.set_angles(positions)

        time.sleep(0.1)



import os
from ..io import get_data_path, log, import_module, get_package_path
from grip_assets import get_data_path
from grip.robot import BulletRobot
from ..robot import (
    BulletPandaGripper,
)

gd = import_module("grip_data")

base_dir = get_data_path() if gd is None else gd.get_data_path()

PANDA_WITH_DEFAULT_GRIPPER_URDF = os.path.join(
    get_package_path("pybullet_data"), "franka_panda/panda.urdf"
)
PANDA_WITH_ROBOTIQ_2F_85_GRIPPER_URDF = os.path.join(
    base_dir, "urdf/franka_panda/panda_with_robotiq_2f_85_gripper.urdf"
)
PANDA_WITH_ROBOTIQ_2F_140_GRIPPER_URDF = os.path.join(
    base_dir, "urdf/franka_panda/panda_with_robotiq_2f_140_gripper.urdf"
)

gripper_type_dict = {
    "default": (PANDA_WITH_DEFAULT_GRIPPER_URDF, BulletPandaGripper, 11),
    "none": ("", None, -1),
}


class BulletPandaArmGripper(BulletRobot):
    def __init__(self, **kwargs):
        self.bullet_gripper_class = kwargs.get("bullet_gripper_class", None)
        self.gripper_type = kwargs.get("gripper_type", "default")

        default_ee_index = -1

        if self.bullet_gripper_class is None:
            if self.gripper_type not in gripper_type_dict.keys():
                log.info("Gripper type {} is None supported!".format(self.gripper_type))
                self.gripper_type = "default"
            gripper_class = gripper_type_dict[self.gripper_type][1]

            if gripper_class is not None:
                kwargs["urdf_file"] = kwargs.get(
                    "urdf_file", gripper_type_dict[self.gripper_type][0]
                )
                default_ee_index = gripper_type_dict[self.gripper_type][2]

        else:
            kwargs["urdf_file"] = kwargs.get("urdf_file", None)
            gripper_class = self.bullet_gripper_class

        self.has_gripper = gripper_class is not None

        kwargs["ee_index"] = kwargs.get("ee_index", default_ee_index)
        kwargs["tip_link"] = kwargs.get("tip_link", "panda_link8")

        kwargs["joint_names"] = kwargs.get(
            "joint_names", ["panda_joint{}".format(idx) for idx in range(1, 8)]
        )

        kwargs["gripper_max_force"] = kwargs.get("gripper_max_force", 12)

        super().__init__(**kwargs)

        self.gripper = None

        if self.has_gripper:
            log.debug("Gripper class {}".format(gripper_class))
            self.gripper = gripper_class(
                parent=self,
                max_force=kwargs["gripper_max_force"],
                ee_index=self.ee_index,
                urdf_file=None,
                sliders=self._sliders,
            )
            self.gripper.control_fingers(mode="open")

    def state(self):
        state = BulletRobot.state(self)

        if self.gripper is not None:
            state["gripper_state"] = self.gripper.state()

        return state

    def reset(self):
        super().reset()

        self.gripper.reset()



import pybullet as p
import numpy as np
from typing import Tuple, Optional
import open3d as o3d

from .sensor import Sensor
from ..perception import RGBD2PCDParser
from ..robot.gui import addDebugFrame, draw_frustum
from ..math import position_quaternion2matrix, matrix2position_quaternion
from ..environments import LightingRandomiser
import math

VISION_META = {
    "view_dist": 1.6,
    "view_pitch": -128.8,
    "view_yaw": -176.6,
    "view_target_pos": [0.14, 0.24, 0.37],
    "width": 512,  # 640,
    "height": 512,  # 480,
    "fx": 256,
    "fy": 256,
    "cx": 256,
    "cy": 256,
    "depth_scale": 1,
    "camera_vfov": 1,
    "camera_hfov": 1.211,
    "camera_near": 0.03,
    "camera_far": 2.0,
    "downsample_height": 64,
    "downsample_width": 64,
}


class RGBDCamera(Sensor):
    """
    RGBDCamera sensor. Represents a simulated RGBD camera.

    Args:
        name (str, optional): camera name. Defaults to "CameraSensor"
        position (numpy.ndarray): position of the optical centre of this camera
        orientation (numpy.ndarray): orientation quaternion of the optical centre of this camera
        anchor_robot (grip.robot.BulletRobot, optional): robot that this camera may be anchored/attached to. Defaults to None.
        anchor_link (str, optional): required if anchor_robot is not None. Link name of the anchor_robot that this camera should be attached to.
            Typically a camera link has been already defined in the URDF of the robot, anchor_link can be chosen to be this pre-defined URDF link (camera optical frame)
        cid (int, optional): physics engine unique identifier (e.g. world.id)
        proj_from_intrinsics (bool, optional): if False sets up intrinsics, sets up projection matrix using self.compute_projection_matrix
            otherwise sets projection matrix using self.build_projection_matrix. Defaults to True.
        frame_id (str, optional): frame name id of this camera.

    """

    def __init__(self, **kwargs):
        self.cid = kwargs.get("cid", None)

        self.params = dict()

        # Default parameters
        self.params.update(VISION_META)

        # Kwargs overwrites default parameters
        self.params.update(kwargs)

        self.name = kwargs.get("name", "CameraSensor")

        self.width = self.params["width"]
        self.height = self.params["height"]
        self.ratio = self.width / self.height

        self.near = self.params["camera_near"]
        self.far = self.params["camera_far"]
        self.params["fx"] = self.params["fx"]
        self.params["fy"] = self.params["fy"]
        self.params["cx"] = self.params["cx"]
        self.params["cy"] = self.params["cy"]
        self.params["depth_scale"] = self.params["depth_scale"]
        self.params["camera_vfov"] = self.params["camera_hfov"] * (
            self.height / self.width
        )
        self.params["camera_aspect"] = float(self.width) / self.height

        self.proj_from_intrinsics = kwargs.get("proj_from_intrinsics", True)

        self.position = kwargs.get("position", [0.86925251, -0.00720329, 1.66544737])
        self.orientation = kwargs.get(
            "orientation", [-0.68534092, -0.70721588, 0.12470087, 0.12084375]
        )
        self.lookat_vector = kwargs.get("lookat_vector", [0, 0, 1])
        self.up_vector = kwargs.get("up_vector", [0, 0, 1])

        self.anchor_robot = kwargs.get("anchor_robot", None)
        self.anchor_link = kwargs.get("anchor_link", None)
        self._anchored_camera = False
        self.frame_id = kwargs.get("frame_id", f"{self.name}_link")
        if self.anchor_robot is not None:
            self.frame_id = self.anchor_link
            pos, ori = self.anchor_robot.get_link_pose_by_name(self.anchor_link)
            self.update_view_from_pose(pos, ori)
            self._anchored_camera = True

        self.set_camera_params()

        if not self.proj_from_intrinsics:
            self.setup_instrinsics()

        self.pcd_parser = RGBD2PCDParser(
            width=self.params["width"],
            height=self.params["height"],
            fx=self.params["fx"],
            fy=self.params["fy"],
            cx=self.params["cx"],
            cy=self.params["cy"],
            camera_near=self.params["camera_near"],
            camera_far=self.params["camera_far"],
            depth_scale=self.params["depth_scale"],
        )

        self._frustum_items = None

    def anchor_camera(
        self, anchor_robot: "grip.robot.BulletRobot", anchor_link: str
    ) -> None:
        """Anchors this camera to a given existing link in the anchor_robot.
        The camera will become rigidly attached to that link.
        The camera will look towards the z-axis of the desired link pose.

        Args:

            anchor_robot (grip.robot.BulletRobot): robot to anchor this camera to.
            anchor_link (str): existing link in anchor_robot to attach this camera to.
        """

        self.anchor_robot = anchor_robot
        self.anchor_link = anchor_link
        self.frame_id = self.anchor_link

        anchor_link_pose = self.anchor_robot.get_link_pose_by_name(self.anchor_link)
        self.update_view_from_pose(*anchor_link_pose)
        self._anchored_camera = True

    def setup_instrinsics(self) -> None:
        """
        Sets up this camera intrinsics parameters.
        It internally sets the intrinsic parameters of this camera from the return of self.get_intrsincs().
        """
        width, height, fx, fy, cx, cy = self.get_intrinsics()
        self.params["width"] = width
        self.params["height"] = height
        self.params["fx"] = fx
        self.params["fy"] = fy
        self.params["cx"] = cx
        self.params["cy"] = cy

    def obs(
        self,
        light_randomiser: Optional[LightingRandomiser] = None,
        shadow: Optional[bool] = True,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Retrieves an observation from this camera.

        Args:
            light_randomiser (Optional[LightingRandomiser]): optinal light randomiser. Warning: lighting randomisation uses simple tiny renderer which is CPU based and slow.
            Optional[bool]: whether or not to enable rendering of shadows in rgb image.
        Returns:
            (Tuple[np.ndarray, np.ndarray, np.ndarray]): three ndarrays with shape-(width, height), respectively, rgb, depth and segmentation mask image (ground-truth segmentation).
        """

        lighting_params = LightingRandomiser.DEFAULT_PARAMS
        renderer = p.ER_BULLET_HARDWARE_OPENGL

        if light_randomiser:
            lighting_params = light_randomiser.randomise()
            renderer = p.ER_TINY_RENDERER

        if self._anchored_camera:
            pos, ori = self.anchor_robot.get_link_pose_by_name(self.anchor_link)
            self.update_view_from_pose(pos, ori)

        _, _, rgb_img, z_buffer, seg_img = p.getCameraImage(
            width=self.params["width"],
            height=self.params["height"],
            viewMatrix=self.params["camera_view_matrix"],
            projectionMatrix=self.params["camera_project_matrix"],
            shadow=shadow,
            lightDirection=lighting_params.light_direction,
            lightColor=lighting_params.light_colour,
            lightAmbientCoeff=lighting_params.ambient_coef,
            lightDiffuseCoeff=lighting_params.diffuse_coef,
            lightSpecularCoeff=lighting_params.specular_coef,
            flags=p.ER_SEGMENTATION_MASK_OBJECT_AND_LINKINDEX,
            renderer=renderer,
            physicsClientId=self.cid,
        )

        if isinstance(rgb_img, tuple):
            rgb_img = np.array(rgb_img).reshape(
                (self.params["height"], self.params["width"], 4)
            )
            z_buffer = np.array(z_buffer)

        rgb_image = rgb_img[:, :, :3].astype(np.uint8)

        # opengl z_buffer to metric space (actual depth)
        depth_image = (
            1.0
            * self.params["camera_far"]
            * self.params["camera_near"]
            / (
                self.params["camera_far"]
                - (self.params["camera_far"] - self.params["camera_near"]) * z_buffer
            )
        )
        depth_image = depth_image.astype(np.float32)

        # seg_image = seg_img.astype(np.uint8)
        seg_image = seg_img

        return rgb_image, depth_image, seg_image

    def rgb2bgr(self, rgb_image: np.ndarray):
        """
        Converts rgb image to bgr image format (e.g. opencv format)

        Returns:
            (numpy.ndarray): bgr image
        """
        bgr_image = rgb_image[..., ::-1].copy()

        # import cv2
        # cv2.imshow("image", bgr_image)
        # cv2.waitKey()

        return bgr_image.astype(np.uint8)

    def pcd_from_rgbd(
        self,
        rgb_image: np.ndarray,
        depth_image: np.ndarray,
        seg_image: np.ndarray = None,
        object_id: list = [],
        vis: bool = False,
    ) -> o3d.geometry.PointCloud:
        """
        Returns open3d point cloud from depth image.

        Args:
            rgb_image (numpy.ndarray): rgb image
            depth_image (numpy.ndarray): depth image
            seg_image (numpy.ndarray, optional): segmentation mask image. Defaults to None.
            object_id (list[int], optional): list of object ids to be kept in the depth cloud.
                This can be used to extract point clouds of selected object ids.
            vis (bool, optional): visualisation for debugging. Default is False.
        Returns:
            (open3d.geometry.PointCloud): point cloud with colours
        """
        return self.pcd_parser.object_pcd_from_rgbd(
            rgb_image,
            depth_image,
            seg_image,
            object_id=object_id,
            vis=vis,
            extrinsics=np.linalg.inv(position_quaternion2matrix(*self.pose)),
        )

    def pcd_from_depth(
        self,
        depth_image: np.ndarray,
        seg_image: np.ndarray = None,
        object_id: list = [],
        vis: bool = False,
    ) -> o3d.geometry.PointCloud:
        """
        Returns open3d point cloud from depth image.

        Args:
            depth_image (numpy.ndarray): depth image
            seg_image (numpy.ndarray, optional): segmentation mask image. Defaults to None.
            object_id (list[int], optional): list of object ids to be kept in the depth cloud.
                This can be used to extract point clouds of selected object ids.
            vis (bool, optional): visualisation for debugging. Default is False.
        Returns:
            (open3d.geometry.PointCloud): point cloud without colours
        """
        return self.pcd_parser.object_pcd_from_depth(
            depth_image,
            seg_image,
            object_id=object_id,
            vis=vis,
            extrinsics=self.get_extrinsics(),
        )

    def stitch_pcd(
        self, xyz: np.ndarray, rgb: np.ndarray, vis: bool = False
    ) -> o3d.geometry.PointCloud:
        """
        Stitch pointclouds extracted from multiple views into a single pointcloud.

        Args:
            xyz (list[numpy.ndarray]): list of xyz values for pointclouds from multiple views
            rgb (list[numpy.ndarray]): list of rgb values for pointclouds from multiple views
            vis (bool, optional): visualize the stitched pointcloud
        Returns:
            (open3d.geometry.PointCloud): The stitched single point cloud
        """
        return self.pcd_parser.stitch_pcd(xyz, rgb, vis=vis)

    # Needs testing
    def update_view_from_pose(
        self, position: np.ndarray, quaternion: np.ndarray
    ) -> None:
        """
        Sets pose of this camera.

        Args:
            position (numpy.ndarray): position of the optical centre of this camera
            orientation (numpy.ndarray): orientation quaternion of the optical centre of this camera
        """
        self.position = position
        self.orientation = quaternion

        opengl_view_mat = position_quaternion2matrix(self.position, self.orientation)
        opengl_view_mat[:3, 1] *= -1
        opengl_view_mat[:3, 2] *= -1

        opengl_view_mat = np.linalg.inv(opengl_view_mat).T

        self.params["camera_view_matrix"] = opengl_view_mat.flatten()

        self.lookat_vector = opengl_view_mat[:3, 2]
        self.up_vector = opengl_view_mat[:3, 1]

    def set_lookat(
        self,
        position: np.ndarray,
        lookat_vector: np.ndarray,
        up_vector: np.ndarray,
        debug: bool = False,
    ) -> None:
        """
        Sets pose of this camera based on position, lookat and up vectors.

        Args:
            position (numpy.ndarray): position of the optical centre of this camera
            lookat_vector (numpy.ndarray): direction of gaze, or look at vector.
            up_vector (numpy.ndarray): up vector of this camera
        """
        self.position = position
        self.lookat_vector = lookat_vector
        self.up_vector = up_vector

        self.update_view_from_lookat(debug=debug)

    def update_view_from_lookat(self, debug: bool = False) -> None:
        """
        Update this camera view parameters based on look at parameters.

        Args:
            debug (bool, optional): whether or not debug mode is on. Default is False.
        """
        self.params["camera_view_matrix"] = p.computeViewMatrix(
            self.position, self.lookat_vector, self.up_vector
        )

        self.position, self.orientation = self.pcd_parser.update_extrinsic(
            self.params["camera_view_matrix"], self.position
        )

        if debug:
            addDebugFrame(self.position, self.orientation, duration=100)

    def update_from_debug_view(self):
        self.position, self.orientation = self.get_debug_view_pose()

        self.update_view_from_pose(self.position, self.orientation)

        # addDebugFrame(self.position, self.orientation, duration=5, lw=5)

    def update_intrinsics(self) -> None:
        """
        Update this camera intrinsics parameters.
        """
        # Option (1) update the projection matrix with known hfov and vfov from camera setting
        #            vfov and fhov = 90 degree for pybullet default visualiser
        # Option(2) update projection with known fx, fy, cx, cy which usually come from camera info / calibration result

        if self.proj_from_intrinsics:
            self.params["camera_project_matrix"] = self.build_projection_matrix(
                self.params["width"],
                self.params["height"],
                self.params["fx"],
                self.params["fy"],
                self.params["cx"],
                self.params["cy"],
                self.params["camera_near"],
                self.params["camera_far"],
            )
        else:
            self.params["camera_project_matrix"] = self.compute_projection_matrix()

    def compute_projection_matrix(self) -> np.ndarray:
        """
        Returns projection matrix using camera frustum parameters.

        Returns:
            (numpy.ndarray): camera projection matrix (opengl)

        """

        return p.computeProjectionMatrix(
            left=-math.tan(float(self.params["camera_hfov"]) / 2.0)
            * self.params["camera_near"],
            right=math.tan(float(self.params["camera_hfov"]) / 2.0)
            * self.params["camera_near"],
            bottom=-math.tan(float(self.params["camera_vfov"]) / 2.0)
            * self.params["camera_near"],
            top=math.tan(float(self.params["camera_vfov"]) / 2.0)
            * self.params["camera_near"],
            nearVal=self.params["camera_near"],
            farVal=self.params["camera_far"],
        )

    def build_projection_matrix(
        self,
        width: float,
        height: float,
        fx: float,
        fy: float,
        cx: float,
        cy: float,
        near: float,
        far: float,
    ) -> np.ndarray:
        """
        Returns projection matrix using camera intrinsic parameters.

        Args:
            width (float): image width
            height (float): image height
            fx (float): camera horizontal focal length
            fy (float): camera vertical focal length
            cx (float): camera image centre of projection coordinate x
            cy (float): camera image centre of projection coordinate y
            near (float): camera near clipping plane
            far (float): camera far clipping plane.

        Returns:
            (numpy.ndarray): camera projection matrix (opengl)

        """
        perspective = np.array(
            [
                [fx, 0.0, -cx, 0.0],
                [0.0, -fy, -cy, 0.0],
                [0.0, 0.0, near + far, near * far],
                [0.0, 0.0, -1.0, 0.0],
            ]
        )

        ortho = self.gl_ortho(0.0, width, height, 0.0, near, far)
        proj_matrix = np.matmul(ortho, perspective)
        proj_matrix = proj_matrix.flatten(order="F")
        return proj_matrix

    def gl_ortho(
        self,
        left: float,
        right: float,
        bottom: float,
        top: float,
        near: float,
        far: float,
    ) -> np.ndarray:
        """
        OpenGL orthogonal projection matrix.

        Returns:
            (numpy.ndarray): opengl orthogonal projection matrix
        """
        ortho = np.diag(
            [2.0 / (right - left), 2.0 / (top - bottom), -2.0 / (far - near), 1.0]
        )
        ortho[0, 3] = -(right + left) / (right - left)
        ortho[1, 3] = -(top + bottom) / (top - bottom)
        ortho[2, 3] = -(far + near) / (far - near)
        return ortho

    def set_projection_matrix(self, projection_matrix: np.ndarray) -> None:
        """
        Sets opengl camera projection matrix for this camera

        Args:
            projection_matrix (numpy.ndarray): opengl projection matrix
        """
        self.params["camera_project_matrix"] = projection_matrix

    def update_camera_params(self) -> None:
        """
        Updates camera params using the update_view_from_lookat
        """
        self.update_view_from_lookat()

    def set_camera_params(self) -> None:
        """
        Sets camera params using the update_view_from_pose
        """
        # Update intrinsics
        self.update_intrinsics()

        self.update_view_from_pose(self.position, self.orientation)

    @property
    def pose(self):
        """
        (Tuple[numpy.ndarray, numpy.ndarray]): camera pose, position and quaternion
        """
        return self.position, self.orientation

    def get_debug_view_matrix(self):
        """
        Gets opengl view matrix from debug visualiser

        Returns:
            (numpy.ndarray): opengl view matrix
        """

        camera_info = p.getDebugVisualizerCamera(physicsClientId=self.cid)
        opengl_view_mat = np.array(camera_info[2]).reshape(4, 4)

        return opengl_view_mat

    def get_debug_view_pose(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Gets pose of the debug visualiser camera.

        Returns:
            (Tuple[numpy.ndarray, numpy.ndarray]): camera pose, position and quaternion.
        """
        opengl_view_mat = self.get_debug_view_matrix()

        camera_pose_gl = np.linalg.inv(opengl_view_mat).transpose()
        camera_pose_gl[:3, 1] *= -1
        camera_pose_gl[:3, 2] *= -1

        camera_pos, camera_quat = matrix2position_quaternion(camera_pose_gl)

        return camera_pos, camera_quat

    def get_debug_camera_params(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Gets params of the debug visualiser camera.

        Returns:
            (float, float, float, Tuple[float]): yaw, pitch, dist, target.
            yaw: yaw angle of the camera, in Cartesian local space coordinates
            pitch: pitch angle of the camera, in Cartesian local space coordinates
            dist: distance between the camera and the camera target
            target: target of the camera, in Cartesian world space coordinates

        """

        camera_info = p.getDebugVisualizerCamera(physicsClientId=self.cid)

        return camera_info[8:]

    def get_debug_view_projection_matrix(self) -> np.ndarray:
        """
        Gets projection matrix of the debug visualiser camera.

        Returns:
            (numpy.ndarray): opengl projection matrix
        """

        camera_info = p.getDebugVisualizerCamera(physicsClientId=self.cid)
        projection_mat = np.array(camera_info[3]).reshape(4, 4)

        return projection_mat

    def get_debug_view_image_size(self) -> Tuple[float, float]:
        """
        Gets the width and height parameters of the debug visualiser camera

        Returns:
            (Tuple[float, float]): width and height
        """
        camera_info = p.getDebugVisualizerCamera(physicsClientId=self.cid)

        return camera_info[0], camera_info[1]

    def get_intrinsics(self) -> Tuple[float, float, float, float, float, float]:
        """
        Gets the raw intrinsics parameters of this camera.

        Returns:
            Tuple[float, float, float, float, float, float]: intrinsics parameters width, height, fx, fy, cx, cy
        """

        width = self.params["width"]
        height = self.params["height"]

        proj_mat = np.array(self.params["camera_project_matrix"]).reshape([4, 4])
        fx = proj_mat[0, 0] / 2.0 * width
        fy = proj_mat[1, 1] / 2.0 * height
        cx = (1.0 - proj_mat[2, 0]) / 2.0 * width
        cy = (1.0 + proj_mat[2, 1]) / 2.0 * height

        return width, height, fx, fy, cx, cy

    def get_o3d_intrinsics(self) -> o3d.camera.PinholeCameraIntrinsic:
        """
        Gets the intrinsic parameters of this camera as an open3d PinholeCameraIntrinsic object.

        Returns:
            (open3d.camera.PinholeCameraIntrinsic): intrinsics parameters of this camera
        """
        width, height, fx, fy, cx, cy = self.get_intrinsics()

        camera_intrinsic = o3d.camera.PinholeCameraIntrinsic(
            width=width, height=height, fx=fx, fy=fy, cx=cx, cy=cy
        )

        return camera_intrinsic

    def get_extrinsics(self) -> np.ndarray:
        """
        Gets the exintrics matrix of this camera (the inverse of its pose).

        Returns:
            (numpy.ndarray): camera extrinsics shape-(4,4)
        """

        return np.linalg.inv(position_quaternion2matrix(*self.pose))

    def get_camera_info(self) -> dict:
        """
        Constructs pinhole camera model matrix from opengl normalised device coordinates (NDC) perspective projection matrix.
        Equations for fx, fy, cx and cy are given by reversing the equations from the opengl NDC perspective projection matrix
        An example is given here: https://blog.noctua-software.com/opencv-opengl-projection-matrix.html
        More complete source opengl transforms: http://www.songho.ca/opengl/gl_transform.html

        Returns:
            (dict): a dict with format
            ``dict: {'width': float, 'height': float, 'distortion_model': str, 'D': numpy.ndarray, 'K': numpy.ndarray, 'R': numpy.ndarray, 'P': numpy.ndarray }``
        """

        width, height, fx, fy, cx, cy = self.get_intrinsics()

        d_matrix = list(np.zeros(5))

        view_matrix = list(np.zeros(9))
        view_matrix[0] = fx
        view_matrix[1] = 0.0
        view_matrix[2] = cx
        view_matrix[3] = 0.0
        view_matrix[4] = fy
        view_matrix[5] = cy
        view_matrix[6] = 0.0
        view_matrix[7] = 0.0
        view_matrix[8] = 1.0

        r_matrix = list(np.zeros(9))
        r_matrix[0] = 1.0
        r_matrix[1] = 0.0
        r_matrix[2] = 0.0
        r_matrix[3] = 0.0
        r_matrix[4] = 1.0
        r_matrix[5] = 0.0
        r_matrix[6] = 0.0
        r_matrix[7] = 0.0
        r_matrix[8] = 1.0

        projection_matrix = list(np.zeros(12))
        projection_matrix[0] = fx
        projection_matrix[1] = 0.0
        projection_matrix[2] = cx
        projection_matrix[3] = 0.0
        projection_matrix[4] = 0.0
        projection_matrix[5] = fy
        projection_matrix[6] = cy
        projection_matrix[7] = 0.0
        projection_matrix[8] = 0.0
        projection_matrix[9] = 0.0
        projection_matrix[10] = 1.0
        projection_matrix[11] = 0.0

        camera_info_dict = {
            "width": width,
            "height": height,
            "distortion_model": "plumb_bob",
            "d": d_matrix,
            "k": view_matrix,
            "r": r_matrix,
            "p": projection_matrix,
        }

        return camera_info_dict

    def draw_frustum(self, duration: float = 0.2) -> None:
        """
        Draws the frustom of this camera in the debug visualiser.

        Args:
            duration (float, optional): lifetime duration of this frustum. Defaults to 0.2s.
        """
        self._frustum_items = draw_frustum(
            self.position,
            self.orientation,
            self.params["camera_hfov"],
            self.params["camera_vfov"],
            self.params["camera_near"],
            self.params["camera_far"],
            self.ratio,
            duration=duration,
            cid=self.cid,
            item_ids=self._frustum_items,
        )



import abc


class Entity(abc.ABC):
    """
    Every entity has a state
    """

    @abc.abstractmethod
    def state(self):
        """
        Reads current entity state
        """
        raise NotImplementedError("state: NO EFFECT, NOT IMPLEMENTED")

    @abc.abstractmethod
    def reset(self):
        """
        Restarts this entity to its original state
        """
        raise NotImplementedError("reset: NO EFFECT, NOT IMPLEMENTED")



from .linalg import *
from .quaternion import *
from .transform import *

__all__ = [
    "position_from_matrix",
    "quaternion_from_matrix",
    "quaternion_from_matrix2",
    "quaternion2rotation_matrix",
    "position_quaternion2matrix",
    "matrix2position_quaternion",
    "make_slerp",
    "quaternion_mult",
    "quaternion_diff",
    "quaternion_error",
    "compute_log",
    "exp_map",
    "compute_omg",
    "rpy2quaternion",
    "quaternion2rpy",
    "euler2quaternion",
    "quaternion_conj",
    "axis_angle2quaternion",
    "normalised",
    "as_transform",
    "proj",
    "plane_vector_directors",
    "project_on_plane",
    "multivariate_normal",
    "invert_transformation_matrix",
    "invert_transform",
    "multiply_transform",
    "transform_inv",
    "transform",
]



import abc


class Agent(abc.ABC):
    @abc.abstractmethod
    def observe(self):
        """
        extract raw observation form state
        including postprocess to usable state if necessary
        """
        raise NotImplementedError

    @abc.abstractmethod
    def act(self):
        """
        Acts on the environment
        """
        raise NotImplementedError


class GraspAgent(Agent):
    @abc.abstractmethod
    def postprocess(self, obs):
        """
        postprocessing raw observation to agent input
        """
        raise NotImplementedError

    @abc.abstractmethod
    def call_grasp_generator(self, target_pcd, context_info):
        """
        the method to call corresponding generator service/topic with
        observation point cloud
        """
        raise NotImplementedError

    @abc.abstractmethod
    def parse_grasp_config(self, grasp_config, auxilary_info):
        """
        formatting single solution provided by grasp generator to
        environment pos, quat to execute

        output can be single pose or pregrasp pose + grasping pose
        depend on whether the generator has already provided the
        pregrasp pose
        """
        raise NotImplementedError

    @abc.abstractmethod
    def select_grasp(self, solutions):
        """
        selectting solution with a cartesian planner with obstacles
        to see if both pre-grasp pose and grasping pose are feasible

        [home --> pregrasping pose]: collision free planning
        [pregrasping pose --> grasping pose]: collision ignored
        """
        raise NotImplementedError

    @abc.abstractmethod
    def grasp(self):
        """
        execute grasping according to postprocessed grasping solution
        """
        raise NotImplementedError



import pybullet as p
import numpy as np
from typing import Union, List, Tuple, Dict, TypedDict, Optional

from ..io import log
from ..math import quaternion_error
from .entity import Entity
from .world import BulletWorld
import uuid, os

ObjectState = TypedDict(
    "State", {"pose": Tuple[np.ndarray, np.ndarray], "entity": Entity}
)


class BulletObject(Entity):
    """
    The BulletObject represents an interface to a Bullet rigid body.
    An object of this class exposes several utilities for manipulating rigid bodies.

    Args:
        world (grip.robot.BulletWorld): parent world where the object where the object belongs to.
        urdf_file (str): urdf file path where this object should be loaded from
        id (int, optional): unique identifier of an already existing object (potentially an object already existing in an external physics world server)
            There is no need to pass urdf_file path if id is passed as parameter
        position (numpy.ndarray): shape-(3,) position in the world where the object to be located.
        orientation (numpy.ndarray): shape-(4,) orientation quaternion in the format [x, y, z, w] (JPL format).
        fixed (bool): whether or not the object is anchored/fixed to the world (a fixed object will not budge).
        scale (float): global scaling of the object that allows one to change the object size.
        compliant_dyn (bool): contacts may be less rigid by setting this flag to true (compliant/springy contacts)
    """

    def __init__(self, **kwargs):
        self.uuid = str(uuid.uuid4())

        self._parent_world = kwargs.get("world", None)
        self._urdf_file = kwargs.get("urdf_file", None)

        self._position = kwargs.get("position", [0, 0, -0.3])
        self._orientation = kwargs.get("orientation", [0, 0, 0, 1])
        self._fixed_base = kwargs.get("fixed", True)
        self._scale = kwargs.get("scale", 1.0)
        self._compliant_dyn = kwargs.get("compliant_dyn", False)

        self.phys_id = 0 if self._parent_world is None else self._parent_world.id

        self._mesh = kwargs.get("mesh", None)
        self._mesh_mass = kwargs.get("mesh_mass", 0.2) if not self._fixed_base else 0
        self._mesh_colour = kwargs.get("mesh_colour", None)

        self._flip_normals = kwargs.get("flip_normals", True)

        self._is_concave = kwargs.get("is_concave", False)
        self._force_concave = kwargs.get("force_concave", False)
        self._disable_collision = kwargs.get("disable_collision", False)
        self._with_texture = kwargs.get("with_texture", True)

        self._id = kwargs.get("id", None)

        self._texture_path = None

        flags = 0

        if self._force_concave:
            flags |= p.GEOM_FORCE_CONCAVE_TRIMESH

        # p.URDF_USE_MATERIAL_COLORS_FROM_MTL

        self._vid = None

        if self._id is None and self._urdf_file is not None:
            self._id = p.loadURDF(
                self._urdf_file,
                self._position,
                self._orientation,
                useFixedBase=self._fixed_base,
                globalScaling=self._scale,
                flags=flags,
                physicsClientId=self.phys_id,
            )
        elif self._mesh is not None:
            self._id, self._vid, self._cid = self._init_from_mesh(self._mesh)
            # disable rendering during creation.

        if self._compliant_dyn:
            # p.changeDynamics(self._id, -1, lateralFriction=1.0, physicsClientId=self.phys_id)
            p.changeDynamics(
                self._id,
                -1,
                contactStiffness=5000,
                contactDamping=500,
                physicsClientId=self.phys_id,
            )

        if self._parent_world is not None:
            self.texture_randomiser = kwargs.get(
                "texture_randomiser", self._parent_world.texture_randomiser
            )
        else:
            self.texture_randomiser = kwargs.get("texture_randomiser", None)

        self._cid = None
        self._base_cid = None
        self._max_force = 500

        self.cached_pose = self.pose

        self.link_dict = self._get_link_dict()

    def _get_mesh_data(self, mesh, flip_normals=True, guiding_normal=[0, 0, 1]):
        import trimesh

        n_verts = len(mesh.vertices)
        uvs = [list(np.random.rand(2)) for i in range(n_verts)]
        try:
            if (
                isinstance(mesh.visual, trimesh.visual.texture.TextureVisuals)
                and len(mesh.visual.uv) > 0
            ):
                uvs = mesh.visual.uv
            else:
                unwrapped_mesh = mesh.unwrap()
                uvs = unwrapped_mesh.visual.uv

        except Exception as e:
            log.warning("Unwrap unavailable: {}".format(e))

        vertices = mesh.vertices

        faces = mesh.faces.reshape(
            -1,
        )

        normals = np.array(mesh.vertex_normals)

        mean_normal = np.mean(normals, axis=0)

        if flip_normals and np.dot(mean_normal, guiding_normal) < 0:
            normals *= -1
            faces = faces[::-1]

        return vertices, faces, normals, uvs

    def _init_from_mesh(self, mesh):
        import trimesh

        # p.configureDebugVisualizer(p.COV_ENABLE_GUI, 0, physicsClientId=self.phys_id)
        # p.configureDebugVisualizer(p.COV_ENABLE_RENDERING, 0, physicsClientId=self.phys_id)
        # disable tinyrenderer, software (CPU) renderer, we don't use it here

        shift = [0, 0.0, 0]
        meshScale = [self._scale, self._scale, self._scale]

        log.info("Mesh kind: {}".format(mesh.visual.kind))

        vertices, faces, normals, uvs = self._get_mesh_data(
            mesh, flip_normals=self._flip_normals
        )

        # normals = -normals
        # the visual shape and collision shape can be re-used by all createMultiBody instances (instancing)

        colour = None

        self._tid = None

        if self._mesh_colour is None and isinstance(
            mesh.visual, trimesh.visual.color.ColorVisuals
        ):
            try:
                colour = trimesh.visual.color.to_float(mesh.visual.vertex_colors)
                log.info("Has colours look: {}".format(colour))
            except:
                log.info("Failed to retrieve colour")

        elif (
            self._mesh_colour is None
            and self._with_texture
            and isinstance(mesh.visual, trimesh.visual.texture.TextureVisuals)
        ):
            colour = trimesh.visual.color.to_float(mesh.visual.to_color().vertex_colors)

            self._texture_path = "/tmp/grip_tex{}.jpg".format(self.uuid)
            mesh.visual.material.image.save(self._texture_path)
            self._parent_world.add_texture(self._texture_path)
            self._tid = self._parent_world.texture(self._texture_path)

        if colour is None:
            colour = (
                [0.8, 0.8, 0.2, 1.0] if self._mesh_colour is None else self._mesh_colour
            )

        vid = p.createVisualShape(
            shapeType=p.GEOM_MESH,
            rgbaColor=colour,
            specularColor=[0.7, 0.4, 0],
            visualFramePosition=shift,
            meshScale=meshScale,
            vertices=vertices,
            indices=faces,
            normals=normals,
            uvs=uvs,
            physicsClientId=self.phys_id,
        )
        cid = -1

        if not self._disable_collision:
            if self._is_concave:
                cid = p.createCollisionShape(
                    shapeType=p.GEOM_MESH,
                    vertices=vertices,
                    collisionFramePosition=shift,
                    # flags=p.GEOM_FORCE_CONCAVE_TRIMESH,
                    meshScale=meshScale,
                    physicsClientId=self.phys_id,
                )
            else:
                cid = p.createCollisionShape(
                    shapeType=p.GEOM_MESH,
                    vertices=vertices,
                    collisionFramePosition=shift,
                    # flags=p.GEOM_FORCE_CONCAVE_TRIMESH,
                    meshScale=meshScale,
                    physicsClientId=self.phys_id,
                )

        bid = p.createMultiBody(
            baseMass=self._mesh_mass,
            baseInertialFramePosition=[0, 0, 0],
            baseCollisionShapeIndex=cid,
            baseVisualShapeIndex=vid,
            basePosition=[0, 0, 0],
            useMaximalCoordinates=False,
            physicsClientId=self.phys_id,
        )

        # shape_data = p.getVisualShapeData(bid, flags=p.VISUAL_SHAPE_DATA_TEXTURE_UNIQUE_IDS, physicsClientId=self.phys_id)

        if self._tid is not None:
            p.changeVisualShape(bid, -1, textureUniqueId=self._tid)

        # p.configureDebugVisualizer(p.COV_ENABLE_RENDERING, 1, physicsClientId=self.phys_id)
        # p.configureDebugVisualizer(p.COV_ENABLE_GUI, 1, physicsClientId=self.phys_id)

        return bid, vid, cid

    def __del__(self):
        if self._texture_path is not None:
            log.info("Clearing resources Bullet object {}".format(self.uuid))
            os.system("rm -rf {}".format(self._texture_path))

    def reset_pose(self, pos: np.ndarray, quat: np.ndarray) -> None:
        """
        Sets position and orientation of this object

        Args:
            pos (numpy.ndarray): shape-(3,) array three dimensional position
            quat (numpy.ndarray): shape-(4,) orientation quaternion following JPL format [x, y, z, w], where w is the real part.

        """

        if self._base_cid is not None:
            p.changeConstraint(self._base_cid, pos, quat, maxForce=self._max_force)
        else:
            p.resetBasePositionAndOrientation(
                self._id, pos, quat, physicsClientId=self.phys_id
            )

    def set_pose(self, pos, quat) -> None:
        """
        Sets position and orientation of this object.
        If the object has a constraint, then the constraint position is changed
        Otherwise, the base pose is reset to the chosen pose.

        Args:
            pos (numpy.ndarray): shape-(3,) array three dimensional position
            quat (numpy.ndarray): shape-(4,) orientation quaternion following JPL format [x, y, z, w], where w is the real part.

        """
        if self._cid is None:
            self.reset_pose(pos, quat)

        else:
            p.changeConstraint(self._cid, pos, quat, maxForce=500)

    def set_colour(self, rgba: List[float]) -> None:
        """
        Sets global color of this object

        Args:
            rgba (list[float]): rgba colour of this object

        """
        if self._vid is None:
            self._vid = -1
        p.changeVisualShape(
            self._id, -1, self._vid, rgbaColor=rgba, physicsClientId=self.phys_id
        )

    def set_friction(self, lateral_friction: float) -> None:
        """Sets object lateral (linear) contact friction coefficient

        Args:
            friction (float): lateral (linear) contact friction coefficient
        """

        p.changeDynamics(self.id, -1, lateralFriction=lateral_friction)

    def set_mass(self, mass: float) -> None:
        """Sets object mass in Kg

        Args:
            mass (float): object's mass in Kg
        """
        p.changeDynamics(self.id, -1, mass=mass)

    def get_dimensions(self) -> np.ndarray:
        """
        Gets the dimensions of this object.

        Returns:
            (numpy.ndarray): depends on geometry type: for GEOM_BOX: extents,
            for GEOM_SPHERE dimensions[0] = radius,
            for GEOM_CAPSULE and GEOM_CYLINDER, dimensions[0] = height (length),
            dimensions[1] = radius.
            For GEOM_MESH, dimensions is the scaling factor.


        """
        data = p.getCollisionShapeData(self.id, -1, physicsClientId=self.phys_id)[0]

        return np.array(data[3])

    def random_texture(self) -> None:
        """
        Randomises texture of this object based on pre-loaded world textures

        """

        if self.texture_randomiser is not None:
            self.texture_randomiser.randomise(self)

    def load_texture(self, texture: Union[List[float], str]) -> None:
        """
        Loads and sets the texture for this object. If texture is a rgba colour, then the colour of this object is set instead.


        Args:

        texture (Union[List[float], str]): rgba colour or string file path to a texture image file.

        """

        assert type(texture) in [list, str, np.str_], "Please provide a valid texture"
        if isinstance(texture, list):
            rgbaColor = texture[:]
            if self._vid is None:
                self._vid = -1
            self.set_colour(rgbaColor)
        elif isinstance(texture, str) or isinstance(texture, np.str_):
            texture_fn = texture
            assert os.path.isfile(texture_fn), "Texture file {} doesn't exit".format(
                texture_fn
            )
            texture_id = p.loadTexture(
                os.path.abspath(texture_fn), physicsClientId=self.phys_id
            )
            p.changeVisualShape(
                objectUniqueId=self._id,
                linkIndex=-1,
                textureUniqueId=texture_id,
                physicsClientId=self.phys_id,
            )

    def remove(self) -> None:
        """
        Removes this object from the world.

        """

        p.removeBody(self._id, physicsClientId=self.phys_id)

    @property
    def id(self):
        """
        int: this object's unique identifier
        """
        return self._id

    @property
    def pose(self):
        """
        Tuple[numpy.ndarray, numpy.ndarray]: this object's pose comprised of three-dimentional position and orientation quaternion of this object.


        """

        position, orientation = p.getBasePositionAndOrientation(
            self._id, physicsClientId=self.phys_id
        )

        return np.array(position), np.array(orientation)

    @property
    def base_velocity(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Gets the current base velocity of this robot, as a linear and angular velocity in cartesian world coordinates.

        Returns:
            (Tuple[numpy.ndarray, numpy.ndarray]): tuple containing a linear shape-(3,) and angular shape-(3,) velocities.
        """
        lin, ang = p.getBaseVelocity(self.id, physicsClientId=self.phys_id)

        return np.asarray(lin), np.asarray(ang)

    @property
    def shape_data(self):
        return p.getVisualShapeData(self.id, -1, physicsClientId=self.phys_id)

    def contact_points(self, entity: Entity) -> List[
        Tuple[
            int,
            int,
            int,
            int,
            int,
            List[float],
            List[float],
            List[float],
            float,
            float,
            float,
            List[float],
            float,
            List[float],
        ]
    ]:
        """
        Returns contact points between this object and another entity given as parameter


        Args:

        entity (grip.robot.Entity): entity we would like to query contact existance

        Returns:
            (List[Tuple[int, int, int, int, int, List[float], List[float], List[float], float, float, float, List[float], float, List[float]]]):
                returns list of bullet contact points
                for more details see: https://docs.google.com/document/d/10sXEhzFRSnvFcl3XxNGhnD4N2SedqwdAvK3dsihxVUA/edit#heading=h.cb0co8y2vuvc

        """

        contact_points = p.getContactPoints(
            entity.id, self.id, physicsClientId=self.phys_id
        )

        return contact_points

    def is_contacting(self, entity: Entity) -> bool:
        """
        Simple binary check if this object is contacting another entity or not

        Args:
            entity (grip.robot.Entity):
                entity we would like to query contact existance

        Returns:
            bool: whether or not this object is contacting the given entity

        """

        contacts = self.contact_points(entity)

        return len(contacts) > 0

    def state(self) -> ObjectState:
        """
        Returns the current state of this object


        Returns:
            (ObjectState): dictionary containing the current state of this object.

        """

        state = dict()

        state["pose"] = self.pose
        state["entity"] = self

        return state

    def record_pose(self) -> None:
        """This stores the pose of object when called.  At a later time youc an called self.moved() to see if the object has moved from the stored pose"""
        self.cached_pose = self.pose
        return self.cached_pose

    def moved(self, pos_threshold: float = 0, ang_threshold: float = 0) -> bool:
        """Call this function to check if the object has moved since the last time record_pose() was called

        Args:
            pos_threshold (float, optional): the maximum distance in meters the object is allowed to have moved and still considered static. Defaults to 0.
            ang_threshold (float, optional): the maximum angular distance in radians the object is allowed to have moved and still considered static. Defaults to 0.

        Returns:
            bool: _description_
        """
        return (
            np.linalg.norm(self.cached_pose[0] - self.pose[0]) > pos_threshold
            and quaternion_error(self.cached_pose[1], self.pose[1]) > ang_threshold
        )

    def reset(self) -> None:
        """
        Resets this object pose to the original position and orientation passed in its contructor during its instantiation

        """

        self.reset_pose(self._position, self._orientation)

    def aabb(self, expansion: float = 0.0) -> Tuple[np.ndarray, np.ndarray]:
        """
        The returns the axis-aligned bounding box of this object


        Returns:
            (Tuple[numpy.ndarray, numpy.ndarray]): a tuple of 3-d vectors (aabb_min, aabb_max) representing the minimum and maximum bounds  in world frame.

        """

        aabb = p.getAABB(self.id, -1, physicsClientId=self.phys_id)

        aabb_min = aabb[0]
        aabb_max = aabb[1]

        return np.array(aabb_min) - expansion, np.array(aabb_max) + expansion

    def create_constraint(
        self,
        position: np.ndarray,
        orientation: np.ndarray,
        entity: Entity = None,
        link_name: str = None,
    ) -> None:
        """
        Creates a rigid constraint to which this object is attached to.

        Args:
            position (numpy.ndarray): position of the constraint
            orientation (numpy.ndarray): orientation of the constraint
            entity (grip.robot.Entity, optional): if passed as argument, then position and orientation is relative to a chosen link name of the child entity
            link_name (str, optional): chosen link name relative to which this object will be attached to. Defaults to None, meaning base_link of the child entity.

        """
        if self._cid is not None:
            return

        child_body = -1 if entity is None else entity.id

        child_link = -1
        if entity is not None and link_name is not None:
            child_link = entity.get_link_id(link_name)

        self._cid = p.createConstraint(
            self._id,
            -1,
            child_body,
            child_link,
            p.JOINT_FIXED,
            [0, 0, 0],  # jointAxis
            [0, 0, 0],  # parentFramePosition
            position,  # childFramePosition
            [0, 0, 0, 1],  # parentFrameOrientation
            orientation,  # childFrameOrientation
            physicsClientId=self.phys_id,
        )

    def create_base_constraint(self):
        if self._base_cid is not None:
            return

        self._position_des = [0, 0, 0]
        self._orientation_des = [0, 0, 0, 1]
        ho = p.getQuaternionFromEuler([0, 0, 0])
        self._base_cid = p.createConstraint(
            self._id,
            -1,
            -1,
            -1,
            p.JOINT_FIXED,
            [0, 0, 0],
            [0.0, 0, 0],
            [0, 0, 0],  # [0.500000, 0.300006, 0.700000],
            ho,
            physicsClientId=self.phys_id,
        )

    def remove_base_constraint(self):
        if self._base_cid is None:
            return

        p.removeConstraint(self._base_cid, physicsClientId=self.phys_id)

        self._base_cid = None

    def disable_object_collision(self, object_entity: Entity) -> None:
        """
        Disables collision between this object and another object_entity

        Args:
            object_entity (grip.robot.Entity): entity with which this object will no longer collide with.
        """
        p.setCollisionFilterPair(
            self.id, object_entity.id, -1, -1, 0, physicsClientId=self.phys_id
        )

    def disable_collision(self) -> None:
        """
        Disables collision between this object and everything else

        """
        p.setCollisionFilterGroupMask(self.id, -1, 0, 0, physicsClientId=self.phys_id)

    def remove_constraint(self) -> None:
        """
        Disables collision between this object and everything else

        """
        if self._cid is not None:
            p.removeConstraint(self._cid, physicsClientId=self.phys_id)

        self._cid = None

    def get_link_ids(self, link_names: Optional[list[str]] = None) -> List[int]:
        """
        Gets the list of link IDs of this robot.

        Args:
            link_names (list[str], optional): optional list of link names. If not passed method returns all link_ids of this kinematic chain
        Returns:
            (List[int]): list of unique identifiers the links belonging to this robot, as loaded by the physics engine.
        """

        if link_names:
            # link ids for selected link names
            return [self.link_dict[lname] for lname in link_names]

        return [lid for _, lid in self.link_dict.items()]

    def _get_link_dict(self) -> Dict[str, int]:
        """
        Gets a dictionary mapping link names to link ids for this object

        Returns:
            (Dict[str, int]): a dictionary mapping link names to corresponding link ids
        """
        link_dict = {}

        for i in range(p.getNumJoints(self.id, physicsClientId=self.phys_id)):
            joint_info = p.getJointInfo(self.id, i, physicsClientId=self.phys_id)

            link_name = joint_info[12].decode("utf-8")
            link_dict[link_name] = i

        # explicitly including base link
        link_dict["-1base"] = -1
        return link_dict


def create_box(
    position: np.ndarray,
    orientation: np.ndarray,
    dimensions_half: np.ndarray,
    world: BulletWorld,
    mass: float = 0.0,
    colour: List[float] = [0.5, 0.5, 0.5, 1.0],
    collision_on: bool = True,
) -> BulletObject:
    """
    Creates a box-shaped BulletObject

    Args:
        position (numpy.ndarray): position of the object
        orientation (numpy.ndarray): orientation of the object
        world (grip.robot.BulletWorld): physics world the created object should belong to.
        mass (float, optional): mass of the object, default is 0.0, meaning it's a static fixed object. Set mass > 0 for dynamic object.
        colour (list[float]): rgba colour of the object.
        collision_on (bool): whether or not this object should have a collision shape.
    """

    shape_id = p.createVisualShape(
        p.GEOM_BOX,
        halfExtents=dimensions_half,
        rgbaColor=colour,
        physicsClientId=world.id,
    )

    box_id = -1
    if collision_on:
        collision_id = p.createCollisionShape(
            p.GEOM_BOX, halfExtents=dimensions_half, physicsClientId=world.id
        )
        box_id = p.createMultiBody(
            mass,
            collision_id,
            shape_id,
            position,
            orientation,
            physicsClientId=world.id,
        )
    else:
        box_id = p.createMultiBody(
            mass, -1, shape_id, position, orientation, physicsClientId=world.id
        )

    obj = BulletObject(
        id=box_id,
        world=world,
        position=position,
        orientation=orientation,
        mesh_colour=colour,
    )
    obj._vid = shape_id

    return obj


def load_sdf(
    filename: str, world: BulletWorld = None, make_z_up: bool = False
) -> List[BulletObject]:
    """Loads a SDF file, returning the list of objects defined within it.
    Which can be later manipulated individually as needed.

    Args:
        filename (str): a relative or absolute path to the SDF file on the file system of the physics server
        world (BulletWorld, optional): world where the sdf should be spawned at. Defaults to None, meaning the default world.

    Returns:
        List[BulletObject]: list of objects loaded from sdf file
    """

    uid_list = p.loadSDF(filename)

    # SDFs may be defined with y-up convention.
    # The default coordinate frame in Bullet is z-up
    # Therefore, if the z-up convention is preferred, make_z_up = True will rotate the entities accordingly.
    if make_z_up:
        for ob_id in uid_list:
            p.resetBasePositionAndOrientation(
                ob_id,
                [0.0, 0.0, 0.0],
                [np.sin(np.pi * 0.25), 0, 0, np.cos(np.pi * 0.25)],
            )

    object_list = [BulletObject(id=uid, world=world) for uid in uid_list]

    return object_list



import abc


class GripApp:
    @abc.abstractmethod
    def setup(self, **kwargs):
        """
        Not implemented
        """
        raise NotImplementedError("setup: NO EFFECT, NOT IMPLEMENTED")

    @abc.abstractmethod
    def run(self, **kwargs):
        """
        Not implemented
        """
        raise NotImplementedError("init: NO EFFECT, NOT IMPLEMENTED")



import os
from typing import Optional, List, Tuple
from ament_index_python.packages import get_package_share_directory
from ..io import pushd, get_file_path, log
import subprocess
import shlex
from .ros2_future import wait_for_message


def xacro_to_urdf_path(
    ros_pkg: str,
    xacro_filename_in: str,
    urdf_filename_out: str,
    xacro_args: Optional[str] = "",
    pkg_subfolder: Optional[str] = "",
    dependencies=[],
) -> str:
    """
    Args:
        ros_pkg: ros package name where the xacro file is located.
        xacro_file_in: xacro filename to look for.
        urdf_filename_out: desired urdf filename output.
        pkg_subfolder: optional package subfolder to inform where to look for xacro_file_in.
    Returns:
        str: full path to output urdf file

    """

    pkg_path = os.path.join(get_package_share_directory(ros_pkg), pkg_subfolder)

    log.debug(f"ROS package path: {pkg_path}")

    xacro_file_path = get_file_path(xacro_filename_in, pkg_path)

    xacro_file_directory = os.path.dirname(xacro_file_path)

    log.debug(f"Xacro file directory: {xacro_file_directory}")

    log.debug(f"Xacro file path: {xacro_file_path}")

    urdf_file_path = os.path.join(xacro_file_directory, urdf_filename_out)

    log.debug(f"Output urdf file path: {xacro_file_path}")

    file_path_out = ""
    with pushd(xacro_file_directory):
        os.system(f"rm {urdf_filename_out} || true 2>&1")
        os.system(
            f"ros2 run xacro xacro {xacro_filename_in} {xacro_args} > {urdf_filename_out}"
        )

        if urdf_filename_out.startswith("/"):
            dir_path = os.path.dirname(urdf_filename_out)
            tmp_pkg_path = f"{dir_path}/{ros_pkg}"
            os.system(f"rm -rf {tmp_pkg_path} || true 2>&1")

            os.system(f"mkdir -p {tmp_pkg_path} || true 2>&1")
            os.system(f"cp -r {pkg_path}/../* {tmp_pkg_path}/")
            os.system(f"cp -r {urdf_filename_out} {tmp_pkg_path}/{pkg_subfolder}")
            _, filename = os.path.split(urdf_filename_out)
            file_path_out = f"{tmp_pkg_path}/{pkg_subfolder}/{filename}"

            for dep in dependencies:
                dp = get_package_share_directory(dep)
                os.system(f"cp -r {dp} {tmp_pkg_path}/{pkg_subfolder}")
        else:
            file_path_out = urdf_file_path

    return file_path_out


def xacro_to_urdf_string(
    ros_pkg: str,
    xacro_filename_in: str,
    xacro_args: Optional[str] = "",
    pkg_subfolder: Optional[str] = "",
) -> str:
    """
    Args:
        ros_pkg: ros package name where the xacro file is located.
        xacro_file_in: xacro filename to look for.
        pkg_subfolder: optional package subfolder to inform where to look for xacro_file_in.
    Returns:
        str: urdf string

    """

    pkg_path = os.path.join(get_package_share_directory(ros_pkg), pkg_subfolder)

    log.debug(f"ROS package path: {pkg_path}")

    xacro_file_path = get_file_path(xacro_filename_in, pkg_path)

    xacro_file_directory = os.path.dirname(xacro_file_path)

    log.debug(f"Xacro file directory: {xacro_file_directory}")

    log.debug(f"Xacro file path: {xacro_file_path}")

    args = shlex.split(f"ros2 run xacro xacro {xacro_filename_in} {xacro_args}")

    sp = subprocess.Popen(
        args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=xacro_file_directory
    )

    output, error = sp.communicate()

    assert len(error) == 0, "Could not convert xacro file to URDF"

    return output.decode("utf-8")


def list_topics(parent_node) -> List[Tuple[str, List[str]]]:
    """Lists existing ros2 topics and corresponding type names

    Returns:
        parent_node: associated parent node
        List[str, List[str]]: list of topic names and associated type names
    """
    from ros2topic.api import get_topic_names_and_types

    topic_names_and_types = get_topic_names_and_types(
        node=parent_node, include_hidden_topics=True
    )

    return topic_names_and_types


def topic_exists(parent_node: "rclpy.node.Node", topic_name: str) -> bool:
    """checks if a given topic exists with topic_name

    Args:
        parent_node: associated parent node
        topic_name (str): the name of to topic to check its existence

    Returns:
        bool: whether or not a topic with given topic name exists
    """

    topic_dict = dict(list_topics(parent_node))

    return topic_name in topic_dict


def is_sim(
    parent_node: "rclpy.node.Node",
    query_topic_test: str = "/joint_states",
    time_to_wait: float = 1,
) -> bool:
    """Checks if we are running in sim or if a real robot already exists in the network

    Naive approach (TODO: implement a less naive approach)
    Assumption: if a robot is already running, then there should already be a /joint_states publishing topic.
    Solution: checks if the topic /joint_states is being published.

    Args:
        parent_node: associated parent node
        query_topic_test: query topic name to be used as reference check
        time_to_wait: time to wait for topic message
    Returns:
        bool: whether or not we are running
    """
    from sensor_msgs.msg import JointState

    received = False

    joint_state_topic_exists = topic_exists(parent_node, query_topic_test)

    if joint_state_topic_exists:
        received, _ = wait_for_message(
            JointState, parent_node, query_topic_test, time_to_wait
        )

    return not (joint_state_topic_exists and received)


def is_real_robot_connected(parent_node: "rclpy.node.Node") -> bool:
    """Checks if we are running in sim or if a real robot already exists in the network

    Naive approach (TODO: implement a less naive approach)
    Assumption: if a robot is already running, then there should already be a /diagnostics topic.
    Solution: checks if the topic /diagnostics is being published.

    Args:
        parent_node: associated parent node
    Returns:
        bool: whether or not we are running
    """

    query_topic_test = "/dynamic_joint_states"
    return topic_exists(parent_node, query_topic_test)



#!/usr/bin/env python3

import os, random
import numpy as np


import pybullet as p

import grip
from .base_env import Env
from .. import robot as g
from ..sensors import RGBDCamera
from ..io import get_data_path
import pybullet_data as pd

base_dir = get_data_path()

TABLE_URDF = os.path.join(base_dir, "urdf/workcell/table/model.urdf")
PLANE_URDF = os.path.join(pd.getDataPath(), "plane.urdf")


VISION_META = {
    "view_dist": 1.6,
    "view_pitch": -128.8,
    "view_yaw": -176.6,
    "view_target_pos": [0.14, 0.24, 0.37],
    "downsample_height": 64,
    "downsample_width": 64,
}


class TemplateEnvironment(Env):
    def __init__(self, **kwargs):
        seed = kwargs.get("seed", 77)

        random.seed(seed)
        np.random.seed(seed)

        self.gripper_type = kwargs.get("gripper", "phase4")
        self.gripper_max_force = kwargs.get("gripper_max_force", 15)

        self.time_step = kwargs.get("time_step", 1.0 / 240.0)
        self.sim_steps = kwargs.get("sim_steps", 10)
        self._object_type = kwargs.get("object_type", ["ycb"])
        self._table_base_pos = kwargs.get("table_base_pos", [0, 0, 0])
        self._render_mode = kwargs.get("render_mode", "GUI")
        self._vision_params = kwargs.get("vision_params", VISION_META)

        self._enable_tex_randomiser = kwargs.get("enable_texture_randomiser", False)

        self.camera_mode = kwargs.get("camera_mode", "on_shoulder")

        self._plane_urdf = PLANE_URDF
        self._table_urdf = TABLE_URDF

        self.entities = dict()
        self.sensors = dict()

        self.world = g.BulletWorld(
            phys_opt=self._render_mode, real_time_sim=False, timestep=self.time_step
        )

        if self._enable_tex_randomiser:
            self.textures_paths = grip.io.file_list(
                grip.io.get_package_path("grip_assets", "texture"),
                extension_filter="jpg",
            )
            self.texture_randomiser = grip.environments.TextureRandomiser(
                self.world, self.textures_paths
            )

        self.rf = None

        self.init()

        self.world.register_keyhandler(ord("r"), self.reset)

    def init(self):
        if self._render_mode == "GUI":
            self.world.setDebugCamera(**self._vision_params)

        # spawn plane and table
        self.plane = g.BulletObject(
            urdf_file=self._plane_urdf, fixed=True, world=self.world
        )

        self.table = g.BulletObject(
            urdf_file=self._table_urdf,
            fixed=True,
            position=self._table_base_pos,
            world=self.world,
        )

        table_p, _ = self.table.pose

        self._table_height = table_p[2]

        # spawn robot
        self.panda = g.BulletPandaArmGripper(
            gripper_type=self.gripper_type,
            gripper_max_force=self.gripper_max_force,
            position=table_p,
            phys_id=self.world.id,
            texture_randomiser=self.world.texture_randomiser,
        )

        self.shoulder_camera = RGBDCamera(
            cid=self.world.id,
            name="ShoulderCamera",
            camera_far=10.0,
            proj_from_intrinsics=False,
        )
        self.shoulder_camera.set_camera_params()
        self.wrist_camera = RGBDCamera(
            cid=self.world.id,
            name="WristCamera",
            anchor_robot=self.panda,
            anchor_link="panda_hand",
        )
        if self.camera_mode in ["on_shoulder", "no_camera"]:
            self.shoulder_camera.draw_frustum(duration=0)

        self.add("plane", self.plane)
        self.add("robot", self.panda)
        self.add("table", self.table)

        if self.camera_mode in ["on_shoulder", "all", "no_camera"]:
            self.add("shoulder_camera", self.shoulder_camera)

        if self.camera_mode in ["on_wrist", "all"]:
            self.add("wrist_camera", self.wrist_camera)

    def get_entity(self, entity_name):
        return self.entities[entity_name]

    def add(self, name, entity):
        if isinstance(entity, grip.robot.Entity):
            self.entities[name] = entity

        if isinstance(entity, grip.sensors.Sensor):
            self.sensors[name] = entity

    def remove(self, name):
        entity = self.entities.pop(name)

        self.world.remove_entity(entity.id)

    def reset(self):
        self.world.restore_state()
        self.panda.reset()

        if self._enable_tex_randomiser:
            self.randomise_textures()

        return self.obs()

    def render(self):
        return self.shoulder_camera.obs()

    def step(self, action):
        if action is None:
            action = {}

        joint_position_cmd = action.get("joint_position_cmd", None)
        gripper_cmd = action.get("gripper_cmd", None)

        if joint_position_cmd is not None:
            self.panda.set_angles(joint_position_cmd)

        if gripper_cmd is not None:
            if isinstance(gripper_cmd, str) and gripper_cmd in ["open", "close"]:
                self.panda.gripper.control_fingers(mode=gripper_cmd)
            else:
                self.panda.gripper.apply_gripper_delta_action(gripper_cmd)

        for _ in range(self.sim_steps):
            self.world.step()

        obs = self.obs()

        reward = 0.0 if self.rf is None else self.rf.reward(self, obs)
        done = False
        info = {}

        return obs, reward, done, info

    def set_reward(self, reward):
        self.rf = reward

    def obs(self):
        obs = dict([(name, e.state()) for (name, e) in self.entities.items()])

        if self.camera_mode != "none":
            for name, camera in self.sensors.items():
                rgb, depth, seg = camera.obs()

                obs[name] = {
                    "rgb": rgb,
                    "depth": depth,
                    "seg": seg,
                    "pose": camera.pose,
                }

        return obs

    def sample(self):
        return {
            "joint_position_cmd": np.random.randn(7),
            "gripper_cmd": np.random.randn(),
        }

    def randomise_textures(self):
        for _, e in self.entities.items():
            e.random_texture()


if __name__ == "__main__":
    env = TemplateEnvironment()

    while p.isConnected():
        env.step(None)



#! /usr/bin/env python
import rclpy
from rclpy.action import ActionServer
from control_msgs.action import FollowJointTrajectory
import grip
from ..motion import Path
import time


class TrajectoryActionServer:
    """Grip Bullet Sim trajectory action server
    Execute trajectory actions

    Args:
        parent_node (rclpy.node.Node): parent node
    """

    def __init__(self, parent_node: "rclpy.node.Node", robot: "grip.robot.BulletRobot"):
        self.node = parent_node
        self.robot = robot

        self.callback_group = rclpy.callback_groups.MutuallyExclusiveCallbackGroup()
        self._action_server = ActionServer(
            self.node,
            FollowJointTrajectory,
            "/joint_trajectory_controller/follow_joint_trajectory",
            self.execute_callback,
            callback_group=self.callback_group,
        )

    def follow_joint_path(
        self, trajectory: FollowJointTrajectory.Goal, dt: float = 0.01
    ):
        if self.robot.sim:
            trajectory = Path.from_ros_trajectory_action_goal(trajectory)

            for wpt in trajectory:
                self.robot.set_angles(wpt, joint_names=trajectory.joint_names)
                time.sleep(dt)

    def execute_callback(
        self, goal_handle: rclpy.action.server.ServerGoalHandle
    ) -> FollowJointTrajectory.Result:
        """Executes trajectory action request

        Returns:
            FollowJointTrajectory.Result: result of trajectory execution
        """
        self.follow_joint_path(goal_handle.request, dt=0.001)
        goal_handle.succeed()
        result = FollowJointTrajectory.Result()
        result.error_code = FollowJointTrajectory.Result.SUCCESSFUL

        return result

    def destroy(self) -> None:
        """Destroy the underlying action server handle."""

        # action server seems to require explicit destroy call (as per 14 March 2023)
        # PRs maybe related to this are currently in rolling distro (not in humble yet):
        # see:
        # https://github.com/ros2/rclpy/pull/1070
        # https://github.com/ros2/rclpy/issues/1034
        self._action_server.destroy()
        self.robot = None
        self.node = None



import abc
from typing import Protocol
from ..io import import_module


class GripperInterface(Protocol):
    # Type name for real hardware interface
    HW_INTERFACE = "path.to.package:ClassName"

    @classmethod
    def get_hw_interface(cls):
        module_path = cls.HW_INTERFACE.split(":")[0]
        class_name = cls.HW_INTERFACE.split(":")[1]

        module = import_module(module_path)

        assert (
            module is not None
        ), f"Importing module {module_path} has failed. Are you sure it exists?"

        return getattr(module, class_name)

    @abc.abstractmethod
    def control_fingers(self, mode: str) -> None:
        """
        Control fingers

        Args:
            mode (str): mode name, e.g. "close", "open"
        """
        raise NotImplementedError("control_fingers: NO EFFECT, NOT IMPLEMENTED")

    @abc.abstractmethod
    def apply_gripper_delta_action(self, action: float) -> None:
        """
        control gripper with a delta finger distance action.

        Args:
            action (float): delta joint angle action for all joints of the gripper
        """
        raise NotImplementedError("control_fingers: NO EFFECT, NOT IMPLEMENTED")

    @abc.abstractproperty
    def angles(self) -> "numpy.ndarray":
        """
        Returns current joint angles measured by encoders

        Returns:
            (numpy.ndarray): joint state of the gripper
        """
        raise NotImplementedError("control_fingers: NO EFFECT, NOT IMPLEMENTED")

    @abc.abstractmethod
    def is_ready(self) -> bool:
        """

        Returns True (bool) when the gripper is ready for operation.

        Returns:
            (Bool): True for gripper state ready

        """
        raise NotImplementedError("is_ready: NO EFFECT, NOT IMPLEMENTED")

    @abc.abstractmethod
    def validate_grasp(self) -> bool:
        """Returns true if the gripper is holding an object

        Returns:
            bool: The gripper has a valid grasp on the object
        """
        raise NotImplementedError(
            "Add method that attempts to close the gripper and check if the gripper is closed"
        )
